class ArithmeticCompilerMixin:
    def FLOAT(self, value: str):
        return float(value)

    def INTEGER(self, value: str):
        return int(value)

    def STRING(self, value: str):
        return value

    def TRUE(self):
        return True

    def FALSE(self):
        return False

    def exp(self, a, b):
        return a ** b

    def neg(self, value):
        return -value

    def div(self, a, b):
        return a / b

    def mul(sef, a, b):
        return a * b

    def mod(self, a, b):
        return a % b

    def add(self, a, b):
        return a + b

    def sub(self, a, b):
        return a - b

    def gt(self, a, b):
        return a > b

    def ge(self, a, b):
        return a >= b

    def lt(self, a, b):
        return a < b

    def le(self, a, b):
        return a <= b

    def eq(self, a, b):
        return a == b

    def ne(self, a, b):
        return a != b

    def IN(self, value, values):
        return value in values

    def NOT(self, value):
        return not value

    def AND(self, a, b):
        return a and b

    def OR(self, a, b):
        return a or b



from dictum_core.utils.expr import (
    value_to_token,
    subselect_column,
    repr_expr_constant,
    column_expr,
)

__all__ = [
    "value_to_token",
    "subselect_column",
    "repr_expr_constant",
    "column_expr",
]



from dictum_core.engine.operators import Operator


class DummyOperator(Operator):
    def __init__(self, input):
        self.input = input
        super().__init__()

    def execute(self, backend):
        if not isinstance(self.input, Operator):
            return self.input
        return self.input.get_result(backend)


def test_run_once():
    op1 = DummyOperator(42)
    op2 = DummyOperator(op1)
    op3 = DummyOperator(op2)
    assert op3.get_result(None) == 42
    assert op3.get_result(None) == 42  # won't call Thread.start() twice



from dictum_core.schema.query import (
    QueryDimension,
    QueryDimensionRequest,
    QueryMetric,
    QueryMetricRequest,
    QueryScalarTransform,
    QueryTableTransform,
    QueryTransform,
)


def test_calculation():
    assert (
        QueryDimensionRequest(dimension=QueryDimension(id="test")).calculation.id
        == "test"
    )
    assert QueryMetricRequest(metric=QueryMetric(id="test")).calculation.id == "test"


def test_calculation_name():
    assert QueryDimension(id="x").name == "x"
    assert (
        QueryDimension(id="x", transforms=[QueryScalarTransform(id="a")]).name == "x__a"
    )
    assert (
        QueryDimension(
            id="x", transforms=[QueryScalarTransform(id="y", args=["a", 1])]
        ).name
        == "x__y_a_1"
    )

    assert QueryMetric(id="x").name == "x"
    assert (
        QueryMetric(
            id="x",
            transforms=[
                QueryTableTransform(
                    id="y",
                    args=[1],
                    of=[QueryDimension(id="a")],
                    within=[
                        QueryDimension(
                            id="b",
                            transforms=[QueryScalarTransform(id="s", args=["ab"])],
                        )
                    ],
                )
            ],
        ).name
        == "x__y_1_of_a_within_b__s_ab"
    )


def test_request_name():
    assert QueryMetricRequest(metric=QueryMetric(id="a"), alias="b").name == "b"
    assert (
        QueryDimensionRequest(dimension=QueryDimension(id="a"), alias="b").name == "b"
    )


def test_render_transform():
    assert QueryTransform(id="test").render() == "test()"
    assert QueryTransform(id="test", args=[1, "a"]).render() == "test(1, 'a')"


def test_render_table_transform():
    assert (
        QueryTableTransform(id="test", args=[1], of=[QueryDimension(id="x")]).render()
        == "test(1) of (x)"
    )
    assert (
        QueryTableTransform(
            id="test",
            args=[1],
            of=[QueryDimension(id="x", transforms=[QueryScalarTransform(id="year")])],
        ).render()
        == "test(1) of (x.year())"
    )
    assert (
        QueryTableTransform(
            id="test",
            args=[1],
            of=[
                QueryDimension(id="x", transforms=[QueryScalarTransform(id="year")]),
                QueryDimension(id="y"),
            ],
        ).render()
        == "test(1) of (x.year(), y)"
    )
    assert (
        QueryTableTransform(
            id="test",
            args=[1],
            of=[
                QueryDimension(id="x", transforms=[QueryScalarTransform(id="year")]),
                QueryDimension(id="y"),
            ],
            within=[
                QueryDimension(id="z"),
            ],
        ).render()
        == "test(1) of (x.year(), y) within (z)"
    )


def test_render_request():
    assert (
        QueryMetricRequest(metric=QueryMetric(id="test"), alias="x").render()
        == 'test as "x"'
    )


def test_render_query_metric():
    assert (
        QueryMetric.parse_obj(
            {"id": "test", "transforms": [{"id": "x"}, {"id": "gt", "args": [1]}]}
        ).render()
        == "test.x().gt(1)"
    )



from hashlib import md5
from typing import List, Optional

from pydantic import BaseModel

from dictum_core import utils


class QueryTransform(BaseModel):
    id: str
    args: List = []

    @property
    def suffix(self):
        args = "_".join(str(a) for a in self.args)
        if len(args) > 0:
            args = f"_{args}"
        return f"{self.id}{args}"

    def render(self) -> str:
        args = ", ".join(utils.repr_expr_constant(a) for a in self.args)
        return f"{self.id}({args})"


class QueryScalarTransform(QueryTransform):
    pass


class QueryTableTransform(QueryTransform):
    of: List["QueryDimension"] = []
    within: List["QueryDimension"] = []

    @property
    def suffix(self) -> str:
        suffix = super().suffix
        for dim in self.of:
            suffix += f"_of_{dim.name}"
        for dim in self.within:
            suffix += f"_within_{dim.name}"
        return suffix

    def render(self) -> str:
        result = super().render()
        if self.of:
            requests = ", ".join(r.render() for r in self.of)
            result += f" of ({requests})"
        if self.within:
            requests = ", ".join(r.render() for r in self.within)
            result += f" within ({requests})"
        return result


class QueryCalculation(BaseModel):
    id: str
    transforms: List[QueryTransform] = []

    @property
    def name(self):
        suffixes = []
        for transform in self.transforms:
            suffixes.append(transform.suffix)
        if len(suffixes) > 0:
            _suff = "_".join(suffixes)
            return f"{self.id}__{_suff}"
        return self.id

    def render(self) -> str:
        result = self.id
        if self.transforms:
            for transform in self.transforms:
                result += f".{transform.render()}"
        return result

    @property
    def digest(self) -> str:
        return md5(self.json(sort_keys=True).encode("UTF-8")).hexdigest()


class QueryCalculationRequest(BaseModel):
    alias: Optional[str]

    @property
    def name(self) -> str:
        if self.alias is not None:
            return self.alias
        return self.calculation.name

    @property
    def digest(self) -> str:
        return self.calculation.digest

    @property
    def calculation(self) -> QueryCalculation:
        raise NotImplementedError

    def render(self) -> str:
        result = self.calculation.render()
        if self.alias:
            result += f' as "{self.alias}"'
        return result


class QueryDimension(QueryCalculation):
    transforms: List[QueryScalarTransform] = []


class QueryDimensionRequest(QueryCalculationRequest):
    dimension: QueryDimension

    @property
    def calculation(self) -> QueryDimension:
        return self.dimension


ops = {
    "eq": "=",
    "ne": "!=",
    "lt": "<",
    "le": "<=",
    "gt": ">",
    "ge": "<",
    "isnull": "is null",
    "isnotnotnull": "is not null",
}


class QueryMetric(QueryCalculation):
    transforms: List[QueryTableTransform] = []

    # def render(self):
    #     """A special case, because no more than 2 are allowed."""
    #     if len(self.transforms) < 2:
    #         return super().render()
    #     table, scalar = self.transforms
    #     result = f"{self.id}.{table.render()}"
    #     op = ops[scalar.id]
    #     val = repr_expr_constant(scalar.args[0]) if scalar.args else ""
    #     result += f" {op} {val}"
    #     return result


class QueryMetricRequest(QueryCalculationRequest):
    metric: QueryMetric
    alias: Optional[str]

    @property
    def calculation(self) -> QueryMetric:
        return self.metric


class Query(BaseModel):
    """A query object that a store understands. Represents a request for metrics
    from a user.
    measures: list of measure identifiers
    dimensions: list of dimension identifiers
    filters: list of boolean expressions. will be concatenated with AND
    """

    metrics: List[QueryMetricRequest] = []
    dimensions: List[QueryDimensionRequest] = []
    filters: List[QueryDimension] = []
    limit: List[QueryMetric] = []

    @property
    def digest(self) -> str:
        return md5(self.json(sort_keys=True).encode("UTF-8")).hexdigest()


QueryTableTransform.update_forward_refs()



from dictum_core.project.altair.format import ldml_number_to_d3_format


def test_sign():
    assert ldml_number_to_d3_format("#") == "d"
    assert ldml_number_to_d3_format("+#") == "+d"
    assert ldml_number_to_d3_format("#;(#)") == "(d"
    assert ldml_number_to_d3_format(" #;-#") == " d"


def test_symbol():
    assert ldml_number_to_d3_format("¤#") == "$d"
    assert ldml_number_to_d3_format("#¤") == "$d"


def test_zero():
    assert ldml_number_to_d3_format("0000") == "04d"
    assert ldml_number_to_d3_format("#,#00") == "02,d"


def test_width():
    assert ldml_number_to_d3_format("00") == "02d"


def test_comma():
    assert ldml_number_to_d3_format("###") == "d"
    assert ldml_number_to_d3_format("#,###") == ",d"
    assert ldml_number_to_d3_format(",##") == ",d"
    assert ldml_number_to_d3_format("#,##,###") == ",d"


def test_precision():
    assert ldml_number_to_d3_format("#.##") == ".2~f"
    assert ldml_number_to_d3_format("#.00") == ".2f"

    # this can't be represented with d3-format
    # we choose between keeping precision (rounding)
    # instead of enforcing zero-padding and having to round to 2
    assert ldml_number_to_d3_format("#.00##") == ".4~f"


def test_type():
    assert ldml_number_to_d3_format("#%") == ".0%"

    # d3-format doesn't support % as prefix, same as #%
    assert ldml_number_to_d3_format("%#") == ".0%"

    assert ldml_number_to_d3_format("#.##E#") == ".2~e"
    assert ldml_number_to_d3_format("#.00E#") == ".2e"

    # in this case babel doesn't format the number as scientific
    # without # or 0 after E, so this is correct
    assert ldml_number_to_d3_format("#.##E") == ".2~f"

    assert ldml_number_to_d3_format("#.") == "d"

    assert ldml_number_to_d3_format("@@@") == ".3r"

    # same as with precision, this can't be represented
    # and we opt for keeping the rounding instead of padding
    assert ldml_number_to_d3_format("@@@##") == ".5~r"



import pytest
from pydantic import BaseModel, ValidationError

from dictum_core.schema.id import ID


class ModelTest(BaseModel):
    id: ID


def test_id_type():
    with pytest.raises(ValidationError, match="must be a string"):
        ModelTest(id=0)


def test_id_identifier():
    with pytest.raises(ValidationError, match="must be a valid Python identifier"):
        ModelTest(id="blah blah")


def test_id_double_underscore():
    with pytest.raises(ValidationError, match="double underscore"):
        ModelTest(id="__blah")


def test_id_uppercase():
    with pytest.raises(ValidationError, match="are reserved"):
        ModelTest(id="Month")


def test_id_ok():
    ModelTest(id="valid_id")



import importlib
from functools import cached_property
from pathlib import Path
from typing import Optional, Union

import altair as alt
import pandas as pd

from dictum_core import schema
from dictum_core.backends.base import Backend
from dictum_core.engine import Engine, Result
from dictum_core.model import Model
from dictum_core.project import analyses
from dictum_core.project.calculations import ProjectDimensions, ProjectMetrics
from dictum_core.project.chart import ProjectChart
from dictum_core.project.magics import QlMagics
from dictum_core.project.templates import environment
from dictum_core.schema import Query


class Project:
    def __init__(
        self,
        path: Optional[Union[str, Path]] = None,
        profile: Optional[str] = None,
    ):
        """
        Arguments:
            path: Path to either the model YAML file or the
                project directory. If a directory, model config is expected to be in
                ``project.yml`` file. Defaults to current working directory.
            profile: Profile name from ``profiles.yml`` to be used. Defaults to the
                default specified in the config.
        """

        if path is None:
            path = Path.cwd()
        if isinstance(path, str):
            path = Path(path)
        if not path.exists():
            raise FileNotFoundError(f"Path {path} does not exist")
        self.path = path
        self.profile = profile
        self.m = ProjectMetrics(self)
        self.metrics = self.m
        self.d = ProjectDimensions(self)
        self.dimensions = self.d
        if self.model.theme is not None:
            alt.themes.register("dictum_theme", lambda: self.model.theme)
            alt.themes.enable("dictum_theme")

    @cached_property
    def _project(self) -> schema.Project:
        return schema.Project.load(self.path)

    @cached_property
    def model(self) -> Model:
        return Model(self._project.get_model())

    @cached_property
    def engine(self) -> Engine:
        return Engine(self.model)

    @cached_property
    def backend(self) -> Backend:
        profile = self._project.get_profile(self.profile)
        return Backend.create(profile.type, profile.parameters)

    def execute(self, query: Query) -> Result:
        computation = self.engine.get_computation(query)
        return computation.execute(self.backend)

    def query_graph(self, query: Query):
        computation = self.engine.get_computation(query)
        return computation.graph()

    def ql(self, query: str):
        return analyses.QlQuery(self, query)

    def select(self, *metrics: str) -> "analyses.Select":
        """
        Select metrics from the project.

        Arguments:
            *metrics: Metric IDs to select.

        Returns:
            A ``Select`` object that can be further modified by chain-calling it's
            methods.
        """
        return analyses.Select(self, *metrics)

    def pivot(self, *metrics: str) -> "analyses.Pivot":
        """Select metrics from the project and construct a pivot table.

        Arguments:
            *metrics: Metric IDs to select.

        Returns:
            A ``Select`` object that can be further modified by chain-calling it's
            methods.
        """
        return analyses.Pivot(self, *metrics)

    @property
    def chart(self) -> ProjectChart:
        return ProjectChart(self)

    @classmethod
    def example(cls, name: str) -> "Project":
        """Load an example project.

        Arguments:
            name (str):
                Name of the example project. Valid values: ``chinook``,
                ``tutorial``.

        Returns:
            CachedProject: same as ``Project``, but won't read the model config at each
            method invocation.
        """
        example = importlib.import_module(f"dictum_core.examples.{name}.generate")
        return example.generate()

    def describe(self) -> pd.DataFrame:
        """Show project's metrics and dimensions and their compatibility. If a metric
        can be used with a dimension, there will be a ``+`` sign at the intersection of
        their respective row and column.

        Returns:
            pandas.DataFrame: metric-dimension compatibility matrix
        """
        print(
            f"Project '{self.model.name}', {len(self.model.metrics)} metrics, "
            f"{len(self.model.dimensions)} dimensions. "
            f"Connected to {self.backend}."
        )
        data = []
        for metric in self.model.metrics.values():
            for dimension in metric.dimensions:
                data.append((metric.id, dimension.id, "✚"))
        return (
            pd.DataFrame(data=data, columns=["metric", "dimension", "check"])
            .pivot(index="dimension", columns="metric", values="check")
            .fillna("")
        )

    def magic(self):
        from IPython import get_ipython  # so that linters don't whine

        ip = get_ipython()
        ip.register_magics(QlMagics(project=self, shell=ip))
        print(
            r"The magic is registered, now you can use %ql and %%ql to query "
            f"{self.model.name} project"
        )

    def _repr_html_(self):
        template = environment.get_template("project.html.j2")
        return template.render(project=self)



import time
from datetime import date, datetime
from typing import Dict, FrozenSet, List, Optional, Union

import dateutil
from lark import Tree
from pandas import DataFrame, concat, isna, merge

from dictum_core import engine
from dictum_core.backends.base import Backend
from dictum_core.backends.pandas import PandasColumnTransformer, PandasCompiler
from dictum_core.engine.computation import Column
from dictum_core.engine.result import ExecutedQuery, Result
from dictum_core.schema import Query


def _date_mapper(v):
    if isinstance(v, date):
        return v
    if isinstance(v, str):
        return dateutil.parser.parse(v).date()
    if isinstance(v, datetime):
        return v.date()
    return v


def _datetime_mapper(v):
    if isinstance(v, datetime):
        return v
    if isinstance(v, date):
        return datetime.combine(v, datetime.min.time())
    if isinstance(v, str):
        return dateutil.parser.parse(v)
    return v


_type_mappers = {
    "bool": bool,
    "float": float,
    "int": int,
    "str": str,
    "date": _date_mapper,
    "datetime": _datetime_mapper,
}


def _set_index(df: DataFrame, columns: List[str]):
    index = list(set(df.columns) & set(columns))
    if index:
        return df.set_index(index)
    return df


class Operator:
    def __init__(self):
        super().__init__()
        self.result = None
        self.ready = False
        self.executed_queries = []

    def execute(self, backend: Backend):
        """Actual operator logic. Execute any upstreams."""
        raise NotImplementedError

    def get_result(self, backend: Backend):
        if self.ready:
            return self.result
        self.result = self.execute(backend)
        self.ready = True
        return self.result

    @property
    def level_of_detail(self) -> List[str]:
        raise NotImplementedError

    @property
    def types(self) -> Dict[str, str]:
        raise NotImplementedError

    @property
    def _dependencies(self) -> List["Operator"]:
        for name in dir(self):
            attr = getattr(self, name)
            if isinstance(attr, Operator):
                yield attr
            elif isinstance(attr, list) and all(isinstance(i, Operator) for i in attr):
                yield from attr

    def walk_graph(self):
        for dep in self._dependencies:
            yield from dep.walk_graph()
        yield self

    def graph(self, *, graph=None, format: str = "png"):
        import graphviz

        self_id = str(id(self))
        self_cls = self.__class__.__name__.replace("Operator", "")

        if graph is None:
            graph = graphviz.Digraph(format=format, strict=True)
            graph.node(self_id, label=self_cls)

        for dependency in self._dependencies:
            dep_id = str(id(dependency))
            graph.node(
                dep_id, label=dependency.__class__.__name__.replace("Operator", "")
            )
            graph.edge(dep_id, self_id)
            dependency.graph(graph=graph)

        graph.graph_attr["rankdir"] = "LR"
        return graph


class MaterializeMixin:
    def materialize(self, inputs: list, backend: Backend) -> List[DataFrame]:
        """Materialize multiple queries on the backend, return a list of DataFrames."""
        results = []
        for input in inputs:  # TODO: parallelize
            if not isinstance(input, DataFrame):
                start = time.time()
                query = backend.display_query(input)
                input = backend.execute(input)
                self.executed_queries.append(
                    ExecutedQuery(query=query, time=time.time() - start)
                )
            results.append(input)
        return results


class QueryOperator(Operator):
    def __init__(self, input: engine.RelationalQuery):
        self.input = input
        self.limit: Optional[int] = None
        self.order: Optional[List[engine.LiteralOrderItem]] = None
        super().__init__()

    def execute(self, backend: Backend):
        self.input.prepare()
        result = backend.compile_query(self.input)
        if self.limit is not None:
            result = backend.limit(result, self.limit)
        if self.order is not None:
            result = backend.order(result, self.order)
        return result

    @property
    def level_of_detail(self) -> List[str]:
        return [c.name for c in self.input._groupby]

    @property
    def types(self) -> Dict[str, str]:
        return {c.name: c.type for c in self.input.columns}


class InnerJoinOperator(Operator, MaterializeMixin):
    def __init__(
        self,
        query: Operator,
        to_join: Operator,
    ):
        super().__init__()
        self.query = query
        self.to_join = to_join

    def execute(self, backend: Backend):
        query = self.query.get_result(backend)
        to_join = self.to_join.get_result(backend)

        if not isinstance(query, DataFrame) and not isinstance(to_join, DataFrame):
            return backend.inner_join(query, to_join, self.level_of_detail)

        # join in pandas
        query, to_join = self.materialize([query, to_join], backend)
        query = _set_index(query, self.level_of_detail)
        to_join = _set_index(to_join, self.level_of_detail)
        return merge(
            query,
            to_join,
            left_index=True,
            right_index=True,
            how="inner",
            suffixes=["", "__joined"],
        )[query.columns].reset_index()

    @property
    def level_of_detail(self) -> List[str]:
        return self.query.level_of_detail

    @property
    def types(self) -> Dict[str, str]:
        return self.query.types


class MergeOperator(Operator, MaterializeMixin):
    def __init__(
        self,
        query: Query,
        inputs: Optional[List[Union[QueryOperator, "MergeOperator"]]] = None,
    ):
        super().__init__()
        self.query = query
        self.inputs = inputs if inputs is not None else []
        self.dimensions: List[Column] = []
        self.metrics: List[Column] = []
        self.limit: Optional[int] = []
        self.order: Optional[List[engine.LiteralOrderItem]] = []

    @property
    def digest(self) -> str:
        return self.query.digest

    @property
    def digests(self) -> FrozenSet[str]:
        return frozenset(i.digest for i in self.inputs if isinstance(i, MergeOperator))

    @property
    def columns(self):
        return self.dimensions + self.metrics

    @property
    def level_of_detail(self) -> List[str]:
        result = set()
        for item in self.inputs:
            result |= set(item.level_of_detail)
        return list(result)

    @property
    def types(self) -> Dict[str, str]:
        return {c.name: c.type for c in self.columns}

    def add_query(self, query: QueryOperator):
        """It's important that dimension columns are added to the merge
        as they appear in the original queries if they are transformed.
        We need to keep the transformed format, name and other display info.
        So, adding these needs to happen automatically, without the engine
        handling getting them out of queries.
        """
        names = {c.name for c in self.dimensions}
        for column in query.input._groupby:
            if column.name not in names:
                self.dimensions.append(
                    Column(
                        name=column.name,
                        expr=Tree("expr", [Tree("column", [None, column.name])]),
                        type=column.type,
                        display_info=column.display_info,
                    )
                )
        self.inputs.append(query)

    def add_merge(self, merge: "MergeOperator"):
        """Add an input merge operator. Checks that the digest doesn't already exist
        in the inputs.
        """
        if merge.digest not in self.digests:
            self.inputs.append(merge)

    def calculate_pandas(self, input: List[list]) -> List[list]:
        column_transformer = PandasColumnTransformer({None: input})
        compiler = PandasCompiler()
        columns = []
        for column in self.columns:
            expr_with_columns = column_transformer.transform(column.expr)
            result = compiler.transformer.transform(expr_with_columns).rename(
                column.name
            )
            columns.append(result)

        result = concat(columns, axis=1)

        if self.order:
            result = result.sort_values(
                by=[i.name for i in self.order],
                ascending=[i.ascending for i in self.order],
            )
        if self.limit:
            result = result.head(self.limit)

        return result

    def merge_pandas(self, inputs: List[DataFrame]) -> DataFrame:
        """Merge multiple DataFrames with Pandas."""
        if len(inputs) == 1:
            return inputs[0]

        result, *rest = inputs
        for df in rest:
            on = list(set(self.level_of_detail) & set(result.columns) & set(df.columns))
            if on:
                result = merge(
                    result,
                    df,
                    left_on=on,
                    right_on=on,
                    how="outer",
                    suffixes=["", "__joined"],
                )
            else:
                result = merge(result, df, how="cross")

        return result

    def merge_queries(self, inputs: list, backend: Backend):
        """Merge multiple queries on the backend."""
        if len(inputs) == 1:
            return inputs[0]
        return backend.merge_queries(inputs, self.level_of_detail)

    def execute(self, backend: Backend) -> List[DataFrame]:
        """Merge multiple inputs. If backend implements merge method and there are
        multiple queries that aren't materialized, use that. If there are any results
        that are DataFrame, materialize the rest and merge in Pandas. Otherwise return
        as is.
        """
        results = []
        for input in self.inputs:
            results.append(input.get_result(backend))

        if len(results) > 1:
            dfs = list(filter(lambda x: isinstance(x, DataFrame), results))
            queries = list(filter(lambda x: not isinstance(x, DataFrame), results))

            result = None
            if len(dfs) > 0:
                materialized = self.materialize(queries, backend)
                result = self.merge_pandas([*dfs, *materialized])
            else:
                try:
                    result = self.merge_queries(queries, backend)
                except NotImplementedError:  # backend doesn't support merge
                    materialized = self.materialize(queries, backend)
                    result = self.merge_pandas(materialized)
        else:
            result = results[0]

        if isinstance(result, DataFrame):
            return self.calculate_pandas(result)

        result = backend.calculate(result, self.columns)
        if self.limit:
            result = backend.limit(result, self.limit)

        # TODO: support order in the query
        if self.order:
            result = backend.order(result, self.order)

        return result


class MaterializeOperator(Operator, MaterializeMixin):
    def __init__(self, inputs: List[Operator]):
        super().__init__()
        self.inputs = inputs

    def execute(self, backend: Backend):
        inputs = [i.get_result(backend) for i in self.inputs]
        return self.materialize(inputs, backend)

    @property
    def level_of_detail(self) -> List[str]:
        result = set()
        for item in self.inputs:
            result |= set(item.level_of_detail)
        return list(result)

    @property
    def types(self) -> Dict[str, str]:
        result = {}
        for item in self.inputs:
            result.update(item.types)
        return result


class FilterOperator(Operator):
    def __init__(self, input: Operator, conditions: List[Tree]):
        self.input = input
        self.conditions = conditions
        super().__init__()

    def execute(self, backend: Backend):
        input = self.input.get_result(backend)
        if not isinstance(input, DataFrame):
            return backend.filter(input, self.conditions)

        # pandas
        column_transformer = PandasColumnTransformer({None: input})
        compiler = PandasCompiler()
        for expr in self.conditions:
            condition = compiler.transformer.transform(
                column_transformer.transform(expr)
            )
            input = input[condition]
        return input

    @property
    def level_of_detail(self) -> List[str]:
        return self.input.level_of_detail

    @property
    def types(self) -> Dict[str, str]:
        return self.input.types


class RecordsFilterOperator(Operator):
    def __init__(
        self,
        query: Operator,
        materialized: MaterializeOperator,
        drop_last_column: bool = False,
    ):
        self.query = query
        self.materialized = materialized
        self.drop_last_column = drop_last_column
        super().__init__()

    def filter_pandas(self, df: DataFrame, filters: List[DataFrame]):
        result = df
        for f in filters:
            columns = list(set(f.columns) & set(result.columns))
            # in case where are duplicate tuples in columns subset
            _f = f[columns].drop_duplicates()
            result = merge(
                result, _f, how="inner", on=columns, suffixes=["", "___drop"]
            )[df.columns]
        return result

    def execute(self, backend: Backend):
        filters: List[DataFrame] = self.materialized.get_result(backend)
        if self.drop_last_column:
            filters = [f.iloc[:, :-1] for f in filters]

        result = self.query.get_result(backend)

        if isinstance(result, DataFrame):
            return self.filter_pandas(result, filters)

        records = [df.to_dict(orient="records") for df in filters]
        return backend.filter_with_records(result, records)

    @property
    def level_of_detail(self) -> List[str]:
        return self.query.level_of_detail

    @property
    def types(self) -> Dict[str, str]:
        return self.query.types


class FinalizeOperator(Operator):
    def __init__(self, input: MaterializeOperator, aliases: Dict[str, str]):
        super().__init__()
        if not isinstance(input, MaterializeOperator):
            raise TypeError(
                "FinalizeOperator expects an instance of MaterializeOperator as input"
            )
        self.input = input
        self.aliases = aliases

    def coerce_types(self, data: List[dict]):
        for row in data:
            for k, v in row.items():
                if row[k] is None:
                    continue
                if isna(row[k]):
                    row[k] = None
                else:
                    row[k] = _type_mappers[self.types[k]](v)
        return data

    def execute(self, backend: Backend):
        """Pluck the materialized df, convert to dicts, coerce types."""
        data = self.input.get_result(backend)[0]
        data = self.coerce_types(
            data.rename(columns=self.aliases).to_dict(orient="records")
        )
        display_info = {}
        for column in self.input.inputs[0].columns:
            info = column.display_info
            info.type = column.type
            if column.name in self.aliases:
                column.name = self.aliases[column.name]
                info.name = column.name
                info.keep_name = True
            display_info[column.name] = info
        return Result(
            data=data,
            executed_queries=self.get_executed_queries(),
            display_info=display_info,
        )

    def get_executed_queries(self) -> List[ExecutedQuery]:
        result = []
        for item in self.walk_graph():
            result.extend(item.executed_queries)
        return result

    @property
    def level_of_detail(self) -> List[str]:
        return self.input.level_of_detail

    @property
    def types(self) -> Dict[str, str]:
        result = {}
        for k, v in self.input.types.items():
            k = self.aliases.get(k, k)
            result[k] = v
        return result



from datetime import datetime
from typing import Any, Dict, List

import numpy as np
import pandas as pd
from lark import Transformer

from dictum_core.backends.base import Compiler
from dictum_core.backends.mixins.arithmetic import ArithmeticCompilerMixin
from dictum_core.backends.mixins.datediff import DatediffCompilerMixin
from dictum_core.engine import Column, LiteralOrderItem, RelationalQuery


class PandasColumnTransformer(Transformer):
    """Replace column references with pd.Series"""

    def __init__(
        self, tables: Dict[str, pd.DataFrame], visit_tokens: bool = True
    ) -> None:
        self._tables = tables
        super().__init__(visit_tokens=visit_tokens)

    def column(self, children: list):
        identity, column = children
        return self._tables[identity][column]


class PandasCompiler(ArithmeticCompilerMixin, DatediffCompilerMixin, Compiler):
    def column(self, table: str, name: str):
        """
        Not required, columns are replaced by pd.Series
        with PandasColumnTransformer
        """

    def IN(self, value, values):
        return value.isin(values)

    def NOT(self, value):
        return ~value

    def AND(self, a, b):
        return a & b

    def OR(self, a, b):
        return a | b

    def isnull(self, value):
        return value.isna()

    def case(self, *whens, else_=None):
        return pd.Series(
            np.select(*zip(*whens), default=else_),
            index=whens[0][0].index,
        )

    # built-in functions
    # aggregate

    def sum(self, arg):
        return arg.sum()

    def avg(self, arg):
        return arg.mean()

    def min(self, arg):
        return arg.min()

    def max(self, arg):
        return arg.max()

    def count(self, args: list):
        """Aggregate count, with optional argument"""
        raise NotImplementedError  # not needed for now

    def countd(self, arg):
        return arg.unique().shape[0]

    # window functions

    def window_sum(self, arg, partition, order, rows):
        if partition:
            return arg.groupby(partition).transform(sum)
        return arg.groupby(pd.Series(0, index=arg.index)).transform(sum)

    def window_row_number(self, args, partition, order, rows):
        if order is None and partition is None:
            return args[0].cumcount() + 1
        if order:
            cols, asc = zip(*(i.children for i in order))
            df = pd.concat([*cols], axis=1)
            df = df.sort_values(by=list(df.columns), ascending=asc)
        if partition == []:
            # create empty groupby
            partition = [pd.Series(data=0, index=df.index)]
        return df.groupby(partition).cumcount() + 1

    # scalar functions

    def abs(self, arg):
        return arg.abs()

    def floor(self, arg):
        return arg.floor()

    def ceil(self, arg):
        return arg.ceil()

    def coalesce(self, *args):
        result, *rest = args
        for item in rest:
            result = result.fillna(item)
        return result

    # type casting

    def tointeger(self, arg):
        return arg.astype(int)

    def tofloat(self, arg):
        return arg.astype(float)

    def todate(self, arg):
        return pd.to_datetime(arg).dt.round("D")

    def todatetime(self, arg):
        return pd.to_datetime(arg)

    # dates

    def datepart(self, part, arg):
        """Part of a date as an integer. First arg is part as a string, e.g. 'month',
        second is date/datetime.
        """
        return getattr(arg.dt, part)

    def datetrunc(self, part, arg):
        """Date truncated to a given part. Args same as datepart."""
        # TODO: support other parts
        mapping = {
            "year": "YS",
            "month": "MS",
            "day": "D",
        }
        return arg.dt.round(mapping[part])

    # for DatediffCompilerMixin
    def datediff_day(self, start, end):
        return (end - start).days

    def now(self):
        return datetime.now()

    def today(self):
        return datetime.today()

    # compilation

    def compile_query(self, query: RelationalQuery):
        raise NotImplementedError

    def merge_queries(self, queries: List, merge_on: List[str]):
        raise NotImplementedError

    def calculate(self, query, columns: List[Column]):
        raise NotImplementedError

    def filter(self, query, conditions: Dict[str, Any]):
        raise NotImplementedError

    def filter_with_records(self, query, records: List[List[Dict[str, Any]]]):
        raise NotImplementedError

    def inner_join(self, query, to_join, join_on: List[str]):
        raise NotImplementedError

    def limit(self, query, limit: int):
        raise NotImplementedError

    def order(self, query, items: List[LiteralOrderItem]):
        raise NotImplementedError



from altair import Chart

import dictum_core.project
from dictum_core.project.altair.data import DictumData
from dictum_core.ql.transformer import compile_dimension, compile_metric


class ProjectChart:
    def __init__(self, project: "dictum_core.project.Project"):
        self.project = project
        self.data = DictumData(project, [])

    def __call__(self, *metrics):
        self.data = DictumData(self.project, metrics)
        return self

    @property
    def chart(self) -> Chart:
        return Chart(self.data)

    def where(self, *filters) -> Chart:
        self.data.filters.extend(compile_dimension(str(f)) for f in filters)
        return self

    def limit(self, *limits) -> Chart:
        self.data.limits.extend(compile_metric(str(lim)) for lim in limits)
        return self

    def __getattr__(self, name: str):
        if name.startswith("_repr_"):
            raise AttributeError
        return getattr(self.chart, name)



import pytest
from lark import Token, Tree

from dictum_core.model import Dimension, Model


@pytest.fixture(scope="module")
def store(store_full):
    return store_full


def test_store_loads(chinook: Model):
    assert chinook.metrics
    assert chinook.measures
    assert chinook.dimensions
    assert chinook.tables


def test_store_parses_expr(chinook: Model):
    assert isinstance(chinook.measures["revenue"].expr, Tree)
    assert isinstance(chinook.dimensions["leap_year"].expr, Tree)
    assert isinstance(chinook.metrics["revenue_per_track"].expr, Tree)


def test_dimension_same_table_as_measures(chinook: Model):
    """There was a bug where the table couldn't find a join path from a self to
    a dimension declared on self :-/
    """
    assert (
        chinook.tables["tracks"].dimension_join_paths.get("track_length_10s_bins")
        is not None
    )


def test_resolve_measures(chinook: Model):
    assert chinook.measures["arppu"].expr == Tree(
        "expr",
        [
            Tree(
                "div",
                [
                    Tree(
                        "call",
                        [
                            "coalesce",
                            Tree(
                                "call",
                                [
                                    "sum",
                                    Tree(
                                        "mul",
                                        [
                                            Tree(
                                                "column", ["invoice_items", "UnitPrice"]
                                            ),
                                            Tree(
                                                "column", ["invoice_items", "Quantity"]
                                            ),
                                        ],
                                    ),
                                ],
                            ),
                            Token("INTEGER", "0"),
                        ],
                    ),
                    Tree(
                        "call",
                        [
                            "countd",
                            Tree("column", ["invoice_items", "invoice", "CustomerId"]),
                        ],
                    ),
                ],
            )
        ],
    )


def test_resolve_metrics(chinook: Model):
    assert chinook.metrics.get("revenue_per_track").expr.children[0] == Tree(
        "div",
        [
            Tree(
                "call",
                ["coalesce", Tree("measure", ["revenue"]), Token("INTEGER", "0")],
            ),
            Tree("measure", ["track_count"]),
        ],
    )


def test_union_is_dimension(chinook: Model):
    union = chinook.tables["customers"].dimensions["country"]
    assert isinstance(union, Dimension)
    assert hasattr(union, "prefixed_expr")


def test_union_allowed(chinook: Model):
    assert "country" in chinook.tables["customers"].dimension_join_paths
    assert "country" in chinook.tables["customers"].allowed_dimensions
    assert "country" not in chinook.tables["invoice_items"].dimension_join_paths
    assert "country" not in chinook.tables["invoice_items"].allowed_dimensions
    assert "customer_country" in chinook.tables["invoice_items"].dimension_join_paths
    assert "customer_country" in chinook.tables["invoice_items"].allowed_dimensions


def test_resolve_aggregate_dimension(chinook: Model):
    assert chinook.dimensions.get("customer_orders_amount").expr.children[0] == Tree(
        "column", ["customers", "__subquery__revenue", "revenue"]
    )


def test_resolve_related_dimension(chinook: Model):
    assert chinook.dimensions.get("order_customer_country").expr.children[0] == Tree(
        "column", ["invoice_items", "invoice", "customer", "Country"]
    )


def test_resolve_dimension_on_anchor(chinook: Model):
    assert chinook.dimensions.get("genre").expr.children[0] == Tree(
        "column", ["genres", "Name"]
    )


def test_resolve_related_aggregate_dimension(chinook: Model):
    assert chinook.dimensions.get("first_order_cohort_month").expr.children[0] == Tree(
        "call",
        [
            "datediff",
            Token("STRING", "month"),
            Tree(
                "column",
                [
                    "invoice_items",
                    "invoice",
                    "customer",
                    "__subquery__min_sale_date",
                    "min_sale_date",
                ],
            ),
            Tree("column", ["invoice_items", "invoice", "InvoiceDate"]),
        ],
    )


def test_inject_default_filters_and_transforms(chinook: Model):
    assert len(chinook.scalar_transforms) == 26


def test_metric_missing(chinook: Model):
    assert chinook.metrics.get("revenue").expr.children[0] == Tree(
        "call", ["coalesce", Tree("measure", ["revenue"]), Token("INTEGER", 0)]
    )



from dictum_core import Project
from dictum_core.project import Select


def test_select(project: Project):
    select = project.select(project.m.revenue)
    assert isinstance(select, Select)
    assert select.df().iloc[0, 0] == 2328.6


def test_select_by(project: Project):
    select = project.select(project.m.revenue).by(project.d.genre)
    assert select.df().shape == (24, 2)


def test_select_where(project: Project):
    select = project.select(project.m.track_count).where(project.d.genre == "Rock")
    assert select.df().iloc[0, 0] == 1297


def test_select_alias(project: Project):
    select = project.select(project.m.revenue).by(
        project.d.invoice_date.year.name("year")
    )
    assert "year" in select.df().columns


def test_select_str(project: Project):
    select = project.select("track_count").by("genre as g").where("genre = 'Rock'")
    df = select.df()
    assert "g" in df.columns
    assert "track_count" in df.columns
    assert df.iloc[0, 1] == 1297



from lark import Transformer, Tree

from dictum_core.ql.parser import (
    parse_dimension,
    parse_dimension_request,
    parse_metric,
    parse_metric_request,
    parse_ql,
)
from dictum_core.schema.query import (
    Query,
    QueryDimension,
    QueryDimensionRequest,
    QueryMetric,
    QueryMetricRequest,
    QueryScalarTransform,
    QueryTableTransform,
)


class QlTransformer(Transformer):
    """Compiles a QL query AST into a Query object."""

    def table_transform(self, children: list):
        id, *rest = children
        args = []
        of, within = [], []
        for item in rest:
            if isinstance(item, Tree):
                if item.data == "of":
                    of = item.children
                elif item.data == "within":
                    within = item.children
            else:
                args.append(item)
        return QueryTableTransform(id=id.lower(), args=args, of=of, within=within)

    def metric(self, children: list):
        id, *transforms = children
        return QueryMetric(id=id, transforms=transforms)

    def scalar_transform(self, children: list):
        id, *args = children
        return QueryScalarTransform(id=id.lower(), args=args)

    def dimension(self, children: list):
        id, *transforms = children
        return QueryDimension(id=id, transforms=transforms)

    def alias(self, children: list):
        return children[0]

    def dimension_request(self, children: list):
        dimension, *rest = children
        alias = rest[0] if rest else None
        return QueryDimensionRequest(dimension=dimension, alias=alias)

    def metric_request(self, children: list):
        metric, *rest = children
        alias = rest[0] if rest else None
        return QueryMetricRequest(metric=metric, alias=alias)

    def select(self, children: list):
        return children

    def query(self, children: list):
        metrics, *rest = children
        filters, dimensions, limit = [], [], []
        for item in rest:
            if item.data == "where":
                filters = item.children
            elif item.data == "groupby":
                dimensions = item.children
            elif item.data == "limit":
                limit = item.children
        return Query(
            metrics=metrics,
            dimensions=dimensions,
            filters=filters,
            limit=limit,
        )


ql_transformer = QlTransformer()


def compile_query(query: str) -> Query:
    return ql_transformer.transform(parse_ql(query))


def compile_dimension(expr: str) -> QueryDimension:
    return ql_transformer.transform(parse_dimension(expr))


def compile_dimension_request(expr: str) -> QueryDimensionRequest:
    return ql_transformer.transform(parse_dimension_request(expr))


def compile_metric(expr: str) -> QueryMetric:
    return ql_transformer.transform(parse_metric(expr))


def compile_metric_request(expr: str) -> QueryMetricRequest:
    return ql_transformer.transform(parse_metric_request(expr))



import pytest

from dictum_core.engine import Engine
from dictum_core.schema import Query


def test_query_validation_metrics(engine: Engine):
    with pytest.raises(ValueError, match="at least one metric"):
        engine.validate_query(Query())


def test_query_validation_duplicate_columns(engine: Engine):
    check = pytest.raises(ValueError, match="Duplicate column name in query")
    with check:
        engine.validate_query(
            Query.parse_obj(
                {"metrics": [{"metric": {"id": "a"}}, {"metric": {"id": "a"}}]}
            )
        )
    with check:
        engine.validate_query(
            Query.parse_obj(
                {
                    "metrics": [
                        {"metric": {"id": "a"}},
                        {"metric": {"id": "b"}, "alias": "a"},
                    ]
                }
            )
        )
    with check:
        engine.validate_query(
            Query.parse_obj(
                {
                    "metrics": [{"metric": {"id": "a"}}],
                    "dimensions": [{"dimension": {"id": "d"}, "alias": "a"}],
                }
            )
        )


def test_query_validation_missing_of_within_dimensions(engine: Engine):
    with pytest.raises(
        ValueError, match="All dimensions used in OF/WITHIN must also be present"
    ):
        engine.validate_query(
            Query.parse_obj(
                {
                    "metrics": [
                        {
                            "metric": {
                                "id": "m",
                                "transforms": [{"id": "t", "of": [{"id": "d"}]}],
                            }
                        }
                    ]
                }
            )
        )
    # correct query is correct
    engine.validate_query(
        Query.parse_obj(
            {
                "metrics": [
                    {
                        "metric": {
                            "id": "m",
                            "transforms": [{"id": "t", "of": [{"id": "d"}]}],
                        }
                    }
                ],
                "dimensions": [{"dimension": {"id": "d"}, "alias": "a"}],
            }
        )
    )



from dictum_core.ql.transformer import (
    compile_dimension,
    compile_dimension_request,
    compile_metric_request,
    compile_query,
)

__all__ = [
    "compile_query",
    "compile_dimension",
    "compile_dimension_request",
    "compile_metric_request",
]



from dictum_core import Project
from dictum_core.project import Pivot


def test_pivot_select(project: Project):
    pivot = project.pivot("revenue", "track_count")
    df = pivot.df()
    assert isinstance(pivot, Pivot)
    assert df.shape == (1, 2)


def test_pivot_rows(project: Project):
    pivot = project.pivot("revenue").rows("genre")
    assert pivot.df().shape == (24, 1)


def test_pivot_columns(project: Project):
    df = project.pivot("revenue").columns("genre").df()
    assert df.shape == (1, 24)
    assert df.index.name is None
    assert df.index == [0]


def test_pivot_both(project: Project):
    pivot = (
        project.pivot("revenue")
        .rows("genre")
        .columns(project.d.invoice_date.year.name("year"))
    )
    df = pivot.df()
    assert df.index.name == "genre"
    assert df.columns.names == ("year", "$")
    assert df.shape == (24, 5)


def test_pivot_measures_reshape(project: Project):
    df = project.pivot("revenue", "track_count").rows("$").columns("genre").df()
    assert df.shape == (2, 25)

    df = (
        project.pivot("revenue", "track_count")
        .rows("$")
        .rows("genre")
        .columns("media_type")
        .df()
    )
    assert df.shape == (50, 5)
    assert df.index.names == ("$", "genre")



from dictum_core.model import Model


def test_measure_dimensions_union(chinook: Model):
    assert "country" in set(
        d.id for d in chinook.measures.get("n_customers").dimensions
    )
    assert "country" in set(d.id for d in chinook.metrics.get("n_customers").dimensions)



from collections import UserDict

from dictum_core.model.calculations import Calculation, Dimension, Measure, Metric


class CalculationDict(UserDict):
    def __getitem__(self, key: str):
        result = self.data.get(key)
        if result is None:
            raise KeyError(f"{self.T.__name__} {key} does not exist")
        return result

    def __setitem__(self, name: str, value):
        if name in self.data:
            raise KeyError(
                f"Duplicate {self.T.__name__.lower()}: {name} on tables "
                f"{self.get(name).table.id} and {value.table.id}"
            )
        return super().__setitem__(name, value)

    def get(self, key: str):
        return self[key]

    def add(self, calc: "Calculation"):
        self[calc.id] = calc


class MeasureDict(CalculationDict):
    T = Measure

    def get(self, key: str) -> Measure:
        return super().get(key)


class DimensionDict(CalculationDict):
    T = Dimension

    def get(self, key: str) -> Dimension:
        return super().get(key)


class MetricDict(CalculationDict):
    T = Metric

    def get(self, key: str) -> Metric:
        return super().get(key)



from typing import List, Union

from pandas import DataFrame

from dictum_core import engine, project
from dictum_core.project.calculations import ProjectDimensionRequest
from dictum_core.project.formatting import Formatter
from dictum_core.ql import (
    compile_dimension,
    compile_dimension_request,
    compile_metric_request,
    compile_query,
)
from dictum_core.schema import Query


class Analysis:
    def __init__(self, project: "project.Project", *metrics):
        self.project = project
        self.query = Query(metrics=[compile_metric_request(str(m)) for m in metrics])

    def _execute(self) -> engine.Result:
        return self.project.execute(self.query)

    def _get_df(self, result: "engine.Result") -> DataFrame:
        return DataFrame(result.data)

    def df(self, *, format: bool = False) -> DataFrame:
        """Execute the query and return a Pandas DataFrame. In Jupyter, if you don't
        need to store the result in a variable and want to just see the data, you can
        skip calling this method. ``Select``'s representation in Jupyter is its result.
        """
        result = self._execute()
        if format:
            return self._get_formatted_df(result)
        return self._get_df(result)

    def _get_formatted_df(self, result: "engine.Result") -> DataFrame:
        if len(result.data) == 0:
            return DataFrame(result.data)
        formatter = Formatter(
            locale=self.project.model.locale,
            formats={k: v.format for k, v in result.display_info.items()},
        )
        df = DataFrame(formatter.format(result.data))
        df.columns = [result.display_info[c].display_name for c in df.columns]
        return df

    def graph(self):
        return self.project.query_graph(self.query)

    @property
    def queries(self) -> List[str]:
        return self.project.execute(self.query).executed_queries

    def _repr_html_(self):
        return self.df(format=True).to_html(max_rows=20)

    def dimensions(self):
        dimensions = self.project.model.suggest_dimensions(self.query)
        return DataFrame(
            data=[
                {"id": d.id, "name": d.name, "description": d.description}
                for d in dimensions
            ]
        )

    def metrics(self):
        metrics = self.project.model.suggest_metrics(self.query)
        return DataFrame(
            data=[
                {"id": m.id, "name": m.name, "description": m.description}
                for m in metrics
            ]
        )


class Select(Analysis):
    """Represents a select query that will return the data in the "tall" format.

    This class is not supposed to be instantiated directly, for instantiating use
    ``Project.select(*metrics: str)``.

    After selecting the metrics, the query can be further refined by calling its
    methods.
    """

    def by(self, *dimensions) -> "Select":
        """Add one or more dimensions to the query. Can be called multiple times for
        adding more than one grouping.

        Arguments:
            dimensions:
                List of one of:

                - Dimension ID in string form to add: ``"date"``
                - String grouping expression with a transform:
                  ``"date.datepart('year')"``
                - Built grouping expression: ``project["date"].datepart("year")``

            alias: An optional alias for the resulting column name. Use this when you
                add the same dimension with different transform multiple times.

        Returns:
            self

        Examples:
            ```py
            (
                project.select("revenue")
                .by("date.year", "year")
                .by("date.month", "month")
            )
            ```

            ```py
            (
                project.select("users_count")
                .by(project["user_revenue"].step(100))
            )
            ```


        """
        for dimension in dimensions:
            if isinstance(dimension, ProjectDimensionRequest):
                request = dimension.request
            else:
                request = compile_dimension_request(str(dimension))
            self.query.dimensions.append(request)
        return self

    def where(self, *filters: str) -> "Select":
        """Add one or more dimension filters to the query. Can be called multiple times.

        Arguments:
            filters: One or more filters to apply.
                One of:

                - String transform expression: ``"amount > 0"``
                - Built transform expression: ``project["amount"] > 0``,
                  ``project["date"].last(30, "day")``

        Returns:
            self

        Examples:
            ```py
            (
                project.select("revenue")
                .by("date.month")
                .where("order_amount >= 10")
            )
            ```
        """
        for f in filters:
            self.query.filters.append(compile_dimension(str(f)))
        return self

    def limit(self, *filters):
        for f in filters:
            self.query.limit.append(compile_metric_request(str(f)).metric)
        return self


class Pivot(Select):
    """Like ``Select``, but allows you to shape the data into an arbitrary _wide_ format.

    This class shouldn't be instantiated directly, use ``Project.pivot`` instead. After
    selecting the metrics, you can further refine and shape the pivot by chain-calling
    its methods.

    There's a special dimension called `$`, which represents metric names. By default
    metric names are displayed at the last level of row index, but you can control this
    by calling either `.rows("$")` or `.columns("$")` in the appropriate order.

    Examples:
        ```py
        (
            project.pivot("revenue")
            .rows("sale_date.year")
            .columns("sale_date.month")
        )
        ```

        ```py
        (
            project.pivot("revenue")
            .rows("channel")
            .columns("sale_date.year")
            .columns("$")
        )
        ```
    """

    def __init__(self, project: "project.Project", *metrics):
        super().__init__(project, *metrics)
        self._rows = []
        self._columns = []

    def rows(self, *dimensions: Union[str, ProjectDimensionRequest]) -> "Pivot":
        """Add a dimension to the row grouping.

        Arguments:
            dimension:
                One of:

                - Dimension ID in string form to add: ``"date"``
                - String grouping expression with a transform:
                  ``"date.datepart('year')"``
                - Built grouping expression: ``project.date.year.name("year")``

            alias: An optional alias for the resulting column name. Use this when you
                add the same dimension with different transform multiple times.

        Returns:
            self
        """
        for dimension in dimensions:
            if isinstance(dimension, str) and dimension == "$":
                self._rows.append(dimension)
            else:
                self.by(dimension)
                dim = self.query.dimensions[-1]
                self._rows.append(dim.name)
        return self

    def columns(self, *dimensions: Union[str, ProjectDimensionRequest]) -> "Pivot":
        """Add a dimension to the column grouping.

        Arguments:
            dimension:
                One of:

                - Dimension ID in string form to add: ``"date"``
                - String grouping expression with a transform:
                  ``"date.datepart('year')"``
                - Built grouping expression: ``project.date.datepart("year")``

            alias: An optional alias for the resulting column name. Use this when you
                add the same dimension with different transform multiple times.

        Returns:
            self
        """
        for dimension in dimensions:
            if isinstance(dimension, str) and dimension == "$":
                self._columns.append(dimension)
            else:
                self.by(dimension)
                dim = self.query.dimensions[-1]
                self._columns.append(dim.name)
        return self

    def _pivot(self, df: DataFrame, rows: List[str], columns: List[str]) -> DataFrame:
        columns = [*columns]
        rows = [*rows]
        df.columns.rename("$", inplace=True)
        if len(rows) == 0 and len(columns) == 0:
            return df

        dimensions = [r.name for r in self.query.dimensions]
        df = df.melt(id_vars=dimensions, value_name="__")

        if "$" not in rows and "$" not in rows:
            columns.append("$")
        if len(rows) == 0:
            df[0] = 0
            rows.append(0)

        df = df.pivot(index=rows, columns=columns, values="__")
        if df.index.nlevels == 1 and df.index.name == 0:
            df.index.name = None

        return df

    def df(self) -> DataFrame:
        df = super().df()
        rows = [*self._rows]
        columns = [*self._columns]
        return self._pivot(df, rows, columns)

    def _repr_html_(self):
        return self.df().to_html(max_rows=20)


class QlQuery(Analysis):
    def __init__(self, project: "project.Project", query: str):
        # not calling super().__init__, because have to parse the query
        self.project = project
        self.query = compile_query(query)



import warnings
from functools import cached_property
from typing import List

from pandas import DataFrame
from sqlalchemy import Integer, String, create_engine
from sqlalchemy.engine.url import URL
from sqlalchemy.exc import SAWarning
from sqlalchemy.sql import Select, case, cast, func

from dictum_core.backends.mixins.datediff import DatediffCompilerMixin
from dictum_core.backends.sql_alchemy import SQLAlchemyBackend, SQLAlchemyCompiler

trunc_modifiers = {
    "year": ["start of year"],
    "month": ["start of month"],
    "week": ["1 day", "weekday 1", "-7 days", "start of day"],
    "day": ["start of day"],
}

trunc_formats = {
    "hour": r"%Y-%m-%d %H:00:00",
    "minute": r"%Y-%m-%d %H:%M:00",
    "second": r"%Y-%m-%d %H:%M:%S",
}

part_formats = {
    "year": r"%Y",
    "month": r"%m",
    "week": r"%W",
    "day": r"%d",
    "hour": r"%H",
    "minute": r"%M",
    "second": r"%S",
}


class SQLiteFunctionsMixin:
    def floor(self, arg):
        return case(
            (arg < 0, cast(arg, Integer) - 1),
            else_=cast(arg, Integer),
        )

    def ceil(self, arg):
        return case(
            (arg > 0, cast(arg, Integer) + 1),
            else_=cast(arg, Integer),
        )

    def div(self, a, b):
        """Fix integer division semantics"""
        return a / self.tofloat(b)

    def todate(self, arg):
        return func.date(arg)

    def todatetime(self, arg):
        return func.datetime(arg)

    def datetrunc(self, part, arg):
        if part in trunc_modifiers:
            modifiers = trunc_modifiers[part]
            return func.datetime(arg, *modifiers)
        if part in trunc_formats:
            fmt = trunc_formats[part]
            return func.strftime(fmt, arg)
        if part == "quarter":
            year = self.datetrunc("year", arg)
            quarter_part = self.datepart_quarter(arg)
            return func.datetime(
                year, "start of year", cast((quarter_part - 1) * 3, String) + " months"
            )
        raise ValueError(
            "Valid values for datetrunc part are year, quarter, "
            "month, week, day, hour, minute, second — "
            f"got '{part}'."
        )

    def datepart_quarter(self, arg):
        return cast(
            (func.strftime("%m", arg) + 2) / 3,
            Integer,
        )

    def datepart_dayofweek(self, arg):
        value = cast(func.strftime("%w", arg), Integer)
        return case({0: 7}, value=value, else_=value)

    def datepart(self, part, arg):
        part = part.lower()
        fmt = part_formats.get(part)
        if fmt is not None:
            return cast(func.strftime(fmt, arg), Integer)
        if part == "quarter":
            return self.datepart_quarter(arg)
        if part in {"dow", "dayofweek"}:
            return self.datepart_dayofweek(arg)
        raise ValueError(
            "Valid values for datepart part are year, quarter, "
            "month, week, day, hour, minute, second — "
            f"got '{part}'."
        )

    def datediff_day(self, start, end):
        start_day = func.julianday(self.datetrunc("day", start))
        end_day = func.julianday(self.datetrunc("day", end))
        return cast(end_day - start_day, Integer)

    def now(self):
        return func.datetime()

    def today(self):
        return func.date()


class SQLiteCompiler(SQLiteFunctionsMixin, DatediffCompilerMixin, SQLAlchemyCompiler):
    pass


class SQLiteBackend(SQLAlchemyBackend):

    type = "sqlite"
    compiler_cls = SQLiteCompiler

    def __init__(self, database: str):
        super().__init__(pool_size=None, default_schema=None, database=database)

    @property
    def url(self) -> str:
        return URL.create(drivername="sqlite", **self.parameters)

    @cached_property
    def engine(self):
        """SQLite doesn't support connection pooling, so have to redefine this"""
        return create_engine(self.url)

    def execute(self, query: Select) -> DataFrame:
        with warnings.catch_warnings():
            warnings.simplefilter("ignore", SAWarning)
            return super().execute(query)

    def merge_queries(self, queries: List[Select], merge_on: List[str]):
        raise NotImplementedError



from dictum_core.schema.model.calculations import Calculation
from dictum_core.schema.model.table import Table
from dictum_core.schema.model.transform import Transform


def test_set_related_ids():
    table = Table.parse_obj(
        {
            "id": "test",
            "source": "test",
            "related": {"test": {"table": "test", "foreign_key": "test"}},
        }
    )
    table.related["test"].alias == "test"


def test_set_measures_ids():
    table = Table.parse_obj(
        {
            "id": "test",
            "source": "test",
            "measures": {"test": {"name": "test", "expr": "test"}},
        }
    )
    table.measures["test"].id == "test"


def test_set_dimensions_ids():
    table = Table.parse_obj(
        {
            "id": "test",
            "source": "test",
            "dimensions": {
                "test": {"name": "test", "expr": "expr", "type": "datetime"}
            },
        }
    )
    table.dimensions["test"].id == "test"


def test_expr_alias():
    calc = Calculation.parse_obj(
        {"id": "test", "name": "test", "type": "int", "expr": "str_expr"}
    )
    assert calc.str_expr == "str_expr"


def test_transform_return_type():
    assert (
        Transform.parse_obj(
            {
                "id": "test",
                "name": "test",
                "expr": "test",
                "return_type": "int",
            }
        ).return_type
        == "int"
    )



from typing import Dict, List, Optional

from babel.dates import format_date, format_datetime, format_skeleton
from babel.numbers import format_currency, format_decimal, format_percent

from dictum_core import schema


class Formatter:
    """Intelligently format the data based on the information in the query."""

    formatters = {
        "number": format_decimal,
        "decimal": format_decimal,
        "percent": format_percent,
        "date": format_date,
        "datetime": format_datetime,
        "skeleton_date": format_skeleton,
        "skeleton_datetime": format_skeleton,
    }

    def __init__(
        self,
        locale: str,
        formats: Dict[str, schema.FormatConfig],
    ):
        self.locale = locale
        self.formats = formats

    def format(self, data: List[dict]):
        for row in data:
            yield self.format_row(row)

    def format_row(self, row: dict):
        return {k: self.format_value(v, self.formats[k]) for k, v in row.items()}

    def format_value(self, value, format: schema.FormatConfig) -> Optional[str]:
        if value is None:
            return ""
        if format.kind == "string":
            return str(value)
        if format.kind == "currency":
            return format_currency(
                value, format.currency, format=format.pattern, locale=self.locale
            )
        if format.skeleton is not None:
            return self.formatters[f"skeleton_{format.kind}"](
                skeleton=format.skeleton, datetime=value, locale=self.locale
            )
        return self.formatters[format.kind](
            value, format=format.pattern, locale=self.locale
        )



import json
import subprocess
import tempfile
from functools import wraps
from pathlib import Path

import altair as alt

altair_output = Path(__file__).parent / "altair_output"


def save_result(fn, name: str):
    @wraps(fn)
    def wrapped(*args, **kwargs):
        result = fn(*args, **kwargs)
        if isinstance(result, alt.TopLevelMixin):
            with tempfile.NamedTemporaryFile(mode="w") as fp:
                json.dump(result._rendered_dict(), fp)
                png = altair_output / f"{name}.png"
                fp.flush()
                subprocess.check_call(
                    [
                        "npx",
                        "-p",
                        "vega",
                        "-p",
                        "vega-lite",
                        "-p",
                        "vega-cli",
                        "vl2png",
                        fp.name,
                        str(png),
                    ]
                )
        return result

    return wrapped


def pytest_collection_modifyitems(session, config, items):
    for item in items:
        item.obj = save_result(item.obj, item.name)



from copy import deepcopy
from dataclasses import dataclass, field
from typing import List, Optional, Union

from lark import Tree

import dictum_core.model
from dictum_core import schema
from dictum_core.engine.result import DisplayInfo
from dictum_core.model import utils
from dictum_core.model.expr import parse_expr


@dataclass
class Column:
    """Represents a column selected from a relational calculation result.

    Arguments:
        name — column name in the resulting query
        expr — Lark expression for the column
        type — column data type
        display_info — info for displaying the column in the formatted table
            or an Altair chart
    """

    name: str
    expr: Tree
    type: schema.Type
    display_info: Optional[DisplayInfo] = None

    @property
    def join_paths(self) -> List[str]:
        result = []
        for ref in self.expr.find_data("column"):
            path = ref.children[1:-1]
            if path:
                result.append(path)
        return result


@dataclass
class Relation:
    source: Union["dictum_core.model.Table", "RelationalQuery"]
    join_tree: List["Join"]

    def add_join_path(self, path: List[str]):
        """Add a single path to the existing join tree. Tables will be joined on the
        relevant foreign key. Path starts with the first related table alias.
        """
        if isinstance(self.source, RelationalQuery):
            raise TypeError(
                "Can't add a join path to a Relation that has a subquery as a source"
            )

        if len(path) == 0:
            return

        alias, *path = path

        # the join path requests a subquery metric join
        # in case of an aggregate dimension
        # inject the necessary subquery and terminate
        if alias.startswith("__subquery__"):
            measure_id = alias.replace("__subquery__", "")
            table = self.source.measure_backlinks.get(measure_id)

            # aggregate table by self.source.primary_key
            measure = table.measures.get(measure_id)
            measure_column = Column(
                name=measure_id,
                expr=measure.expr,
                type=measure.type,
            )
            subquery = RelationalQuery(
                source=table, join_tree=[], _aggregate=[measure_column]
            )
            for join_path in measure_column.join_paths:
                subquery.add_join_path(join_path)
            join_path = table.allowed_join_paths[self.source]
            subquery.add_join_path(join_path)
            pk = Column(
                name="__pk",
                type=None,
                expr=Tree(
                    "expr",
                    [Tree("column", [table.id, *join_path, self.source.primary_key])],
                ),
            )
            subquery.add_groupby(pk)
            join = Join(
                source=subquery,
                join_tree=[],
                alias=alias,
                expr=parse_expr(
                    f"{self.source.id}.{alias}.__pk = "
                    f"{self.source.id}.{self.source.primary_key}"
                ),
            )
            self.join_tree.append(join)
            return

        related = self.source.related[alias]
        join = Join(
            expr=related.join_expr,
            source=related.table,
            alias=alias,
            join_tree=[],
        )

        for existing_join in self.join_tree:
            if join == existing_join:  # if this join already exists, go down a level
                existing_join.add_join_path(path)
                return  # and terminate

        # if this join doesn't exist
        self.join_tree.append(join)
        join.add_join_path(path)  # add and go further down

    @property
    def joins(self) -> List["UnnestedJoin"]:
        return list(self._unnested_joins())

    def _unnested_joins(self, path=()):
        if path == ():
            path = (self.source.id,)
        for join in self.join_tree:
            prefix = (*path, join.alias)
            yield UnnestedJoin(
                expr=utils.prepare_expr(join.expr, path),
                right_identity=".".join(prefix),
                right=join.source,
                inner=join.inner,
            )
            if not isinstance(join.source, RelationalQuery):
                yield from join._unnested_joins(path + (join.alias,))


@dataclass
class Join(Relation):
    alias: str
    expr: Tree
    inner: bool = False

    def prepare(self):
        if isinstance(self.source, RelationalQuery):
            self.source.prepare()
        for join in self.join_tree:
            join.prepare()

    def __eq__(self, other: "Join"):
        return (
            isinstance(other, Join)
            and self.expr == other.expr
            and self.alias == other.alias
            and self.source == other.source
        )


@dataclass
class UnnestedJoin:
    expr: Tree
    right_identity: str
    right: Union["dictum_core.model.Table", "RelationalQuery"]
    inner: bool


@dataclass
class OrderItem:
    expr: Tree
    ascending: bool


@dataclass
class LiteralOrderItem:
    """Like OrderItem, but by name"""

    name: str
    ascending: bool


@dataclass
class RelationalQuery(Relation):
    """
    A simple relational query.

    Arguments:
        _aggregate: A list of aggregate columns.
        _groupby: A list of non-aggregate columns to group by.
        filters: A list of boolean-valued expressions.
        order: A list of OrderItem, each one of which is an expression and a boolean
            telling if the order is ascending or descending.
        limit: An integer limit.
    """

    _aggregate: List[Column] = field(default_factory=list)
    _groupby: List[Column] = field(default_factory=list)
    filters: List[Tree] = field(default_factory=list)
    order: List[OrderItem] = field(default_factory=list)
    limit: Optional[int] = None

    def add_groupby(self, column: Column):
        for path in column.join_paths:
            self.add_join_path(path)
        self._groupby.append(column)

    def add_aggregate(self, column: Column):
        for path in column.join_paths:
            self.add_join_path(path)
        self._aggregate.append(column)

    def add_filter_expr(self, expr: Tree):
        expr = deepcopy(expr)
        for ref in expr.find_data("column"):
            _, *path, _ = ref.children
            self.add_join_path(path)
        self.filters.append(expr)

    @property
    def groupby(self) -> List[Tree]:
        return [c.expr for c in self._groupby]

    @property
    def columns(self) -> List[Column]:
        return [*self._groupby, *self._aggregate]

    @staticmethod
    def prepare_expr(expr: Tree) -> Tree:
        expr = deepcopy(expr)
        for ref in expr.find_data("column"):
            *path, id_ = ref.children
            ref.children = [".".join(path), id_]
        return expr

    def prepare(self):
        for column in self._aggregate + self._groupby:
            column.expr = self.prepare_expr(column.expr)
        self.filters = list(map(self.prepare_expr, self.filters))
        for item in self.order:
            item.expr = self.prepare_expr(item.expr)
        for join in self.join_tree:
            join.prepare()



from abc import ABC, abstractmethod
from typing import Optional

from altair import (
    X2,
    Y2,
    Angle,
    Color,
    Column,
    Detail,
    Facet,
    Fill,
    FillOpacity,
    Href,
    Key,
    Latitude,
    Latitude2,
    Longitude,
    Longitude2,
    Order,
    Radius,
    Row,
    Shape,
    Size,
    Stroke,
    StrokeDash,
    StrokeOpacity,
    StrokeWidth,
    Text,
    Theta,
    Theta2,
    Tooltip,
    X,
    XError,
    XError2,
    Y,
    YError,
    YError2,
)

axis = {
    Latitude,
    Longitude,
    X,
    XError,
    Y,
    YError,
}
legend = {
    Color,
    Angle,
    Fill,
    FillOpacity,
    Radius,
    Shape,
    Size,
    Stroke,
    StrokeDash,
    StrokeOpacity,
    StrokeWidth,
}
header = {Column, Row, Facet}
none = {
    Detail,
    Href,
    Key,
    Text,
    Theta,
    Tooltip,
    Order,
}
field_only = {
    Theta2,
    YError2,
    X2,
    Y2,
    Latitude2,
    Longitude2,
    XError2,
}


class AltairEncodingChannelHook(ABC):
    @abstractmethod
    def encoding_fields(self, cls: Optional[type] = None) -> dict:
        ...


def filter_fields(cls, fields: dict):
    keys = ["field", "type"]
    if cls in axis:
        keys.append("axis")
    if cls in legend:
        keys.append("legend")
    if cls in header:
        keys.append("header")
    if cls in field_only:
        keys.remove("type")
    return {k: fields[k] for k in keys if k in fields}



from typing import List, Optional

from lark import Transformer, Tree

from dictum_core import engine, schema
from dictum_core.model.expr import parse_expr
from dictum_core.utils import value_to_token


class TransformTransformer(Transformer):
    def __init__(self, arg, args: dict, visit_tokens: bool = True) -> None:
        self._arg = arg
        self._args = args
        super().__init__(visit_tokens=visit_tokens)

    def column(self, children: list):
        return self._args[children[0]]

    def ARG(self, _):
        return self._arg

    def expr(self, children):
        return children[0]


transforms = {}


class ScalarTransform:
    """A scalar transform. Column in, Column out. Can change different aspects
    of the column, expression (always), name, format etc.
    """

    id: str
    name: str
    description: Optional[str] = None
    return_type: Optional[schema.Type] = None

    def __init__(self, *args):
        self._args = [value_to_token(a) for a in args]

    def __init_subclass__(cls):
        if hasattr(cls, "id") and cls.id is not None:
            transforms[cls.id] = cls

    def get_name(self, name: str) -> str:
        return name

    def get_display_name(self, name: str) -> str:
        return name

    def get_return_type(self, original: schema.Type) -> schema.Type:
        if self.return_type is not None:
            return self.return_type
        return original

    def get_format(self, format: schema.FormatConfig) -> schema.FormatConfig:
        return format

    def get_display_info(
        self, display_info: Optional["engine.DisplayInfo"]
    ) -> "engine.DisplayInfo":
        if display_info is None:
            return None
        return engine.DisplayInfo(
            display_name=(
                self.get_display_name(display_info.display_name)
                if not display_info.keep_display_name
                else display_info.display_name
            ),
            column_name=display_info.column_name,
            format=self.get_format(display_info.format),
            keep_display_name=display_info.keep_display_name,
            kind=display_info.kind,
        )

    def transform_expr(self, expr: Tree) -> Tree:
        raise NotImplementedError

    def __call__(self, column: "engine.Column"):
        return engine.Column(
            name=self.get_name(column.name),
            type=self.get_return_type(column.type),
            expr=Tree("expr", [self.transform_expr(column.expr.children[0])]),
            display_info=self.get_display_info(column.display_info),
        )


class LiteralTransform(ScalarTransform):
    """A kind of ScalarTransform that's defined with a literal expression.
    @ token gets replaced with expr of the argument.
    """

    expr: str
    args: List[str] = []

    @property
    def _expr(self) -> Tree:
        return parse_expr(self.expr)

    def transform_expr(self, expr: Tree):
        kwargs = dict(zip(self.args, self._args))
        transformer = TransformTransformer(expr, kwargs)
        return transformer.transform(self._expr)


class InvertTransform(LiteralTransform):
    id = "invert"
    return_type = "bool"
    expr = "not (@)"


class BooleanTransform(LiteralTransform):
    args = ["value"]
    return_type = "bool"
    op: str

    def __init_subclass__(cls):
        cls.id = cls.__name__[:2].lower()
        cls.name = cls.op
        cls.expr = f"(@) {cls.op} value"
        super().__init_subclass__()


class EqTransform(BooleanTransform):
    op = "="


class NeTransform(BooleanTransform):
    op = "!="


class GtTransform(BooleanTransform):
    op = ">"


class GeTransform(BooleanTransform):
    op = ">="


class LtTransform(BooleanTransform):
    op = "<"


class LeTransform(BooleanTransform):
    op = "<="


class IsNullTransform(LiteralTransform):
    id = "isnull"
    name = "IS NULL"
    return_type = "bool"
    expr = "@ is null"


class IsNotNullTransform(LiteralTransform):
    id = "isnotnull"
    name = "IS NOT NULL"
    return_type = "bool"
    expr = "@ is not null"


class InRangeTransform(LiteralTransform):
    id = "inrange"
    name = "in range"
    return_type = "bool"
    args = ["min", "max"]
    expr = "@ >= min and @ <= max"


class IsInTransform(ScalarTransform):
    id = "isin"
    name = "IN"
    return_type = "bool"

    def transform_expr(self, expr: Tree) -> Tree:
        return Tree("IN", [expr, *self._args])


class LastTransform(LiteralTransform):
    id = "last"
    name = "last"
    return_type = "bool"
    args = ["n", "part"]
    expr = "datediff(part, @, now()) <= n"

    def __init__(self, n: int, period: str):
        super().__init__(n, period)


class StepTransform(LiteralTransform):
    id = "step"
    name = "step"
    return_type = "int"
    args = ["size"]
    expr = "@ // size * size"


class DatepartTransform(LiteralTransform):
    id = "datepart"
    name = "date part"
    args = ["part"]
    return_type = "int"
    expr = "datepart(part, @)"

    def get_format(self, format: schema.FormatConfig) -> schema.FormatConfig:
        return schema.FormatConfig(kind="decimal", pattern="#")

    def get_display_name(self, name: str) -> str:
        return f"{name} ({self._args[0]})"

    def __init__(self, part: str):
        super().__init__(part)


class ShortDatepartTransform(DatepartTransform):
    id = None
    altair_time_unit = None

    def __init__(self):
        super().__init__(self.id)


class YearTransform(ShortDatepartTransform):
    id = "year"
    name = "Year"


class QuarterTransform(ShortDatepartTransform):
    id = "quarter"
    name = "Quarter"


class MonthTransform(ShortDatepartTransform):
    id = "month"
    name = "Month"


class WeekTransform(ShortDatepartTransform):
    id = "week"
    name = "Week"


class DayTransform(ShortDatepartTransform):
    id = "day"
    name = "Day"


class HourTransform(ShortDatepartTransform):
    id = "hour"
    name = "Hour"


class MinuteTransform(ShortDatepartTransform):
    id = "minute"
    name = "Minute"


class SecondTransform(ShortDatepartTransform):
    id = "second"
    name = "Second"


class DayOfWeekTransform(ShortDatepartTransform):
    id = "dayofweek"
    name = "Day of Week"


class DowTransform(DayOfWeekTransform):
    id = "dow"


date_skeletons = {
    "year": "y",
    "quarter": "yQQQ",
    "month": "yMMM",
    "week": "yw",
    "day": "yMd",
}


class DatetruncTransform(ScalarTransform):
    id = "datetrunc"
    name = "Truncate a date"
    return_type = "datetime"

    part_to_altair_time_unit = {
        "year": "year",
        "quarter": "yearquarter",
        "month": "yearmonth",
        "day": "yearmonthdate",
        "hour": "yearmonthdatehours",
        "minute": "yearmonthdatehoursminutes",
        "second": "yearmonthdatehoursminutesseconds",
    }

    def __init__(self, period: str):
        super().__init__(period)

    def get_format(self, format: schema.FormatConfig) -> schema.FormatConfig:
        part = self._args[0]
        if part in date_skeletons:
            return schema.FormatConfig(kind="date", skeleton=date_skeletons[part])
        return schema.FormatConfig(kind="datetime")

    def get_display_info(
        self, display_info: Optional["engine.DisplayInfo"]
    ) -> "engine.DisplayInfo":
        info = super().get_display_info(display_info)
        info.altair_time_unit = self.part_to_altair_time_unit.get(self._args[0])
        return info

    def transform_expr(self, expr: Tree):
        return Tree("call", ["datetrunc", *self._args, expr])


class DateTransform(DatetruncTransform):
    id = "date"
    name = "Truncate a datetime to a date"

    def __init__(self):
        super().__init__("day")



import datetime
import inspect
import warnings
from itertools import chain
from typing import Dict, List

import pandas as pd
from altair import (
    Column,
    EncodingSortField,
    Facet,
    FacetChart,
    FacetedUnitSpec,
    FieldChannelMixin,
    FieldName,
    LayerChart,
    NormalizedFacetSpec,
    NormalizedSpec,
    RepeatRef,
    RepeatSpec,
    Row,
    SchemaBase,
    Text,
    TopLevelMixin,
    TopLevelRepeatSpec,
    TopLevelUnitSpec,
    Undefined,
    UnitSpec,
    channels,
    renderers,
)
from altair.utils.core import update_nested
from altair.utils.data import to_values
from altair.vegalite.v4.api import _EncodingMixin
from jsonschema.exceptions import ValidationError
from toolz.curried import curry

from dictum_core.engine.computation import DisplayInfo
from dictum_core.project.altair.data import DictumData
from dictum_core.project.altair.encoding import AltairEncodingChannelHook, filter_fields
from dictum_core.project.altair.format import format_config_to_d3_format
from dictum_core.project.altair.locale import (
    cldr_locale_to_d3_number,
    cldr_locale_to_d3_time,
)
from dictum_core.ql.transformer import compile_dimension_request, compile_metric_request
from dictum_core.schema.query import Query, QueryMetricRequest


def is_channel_cls(obj):
    return (
        isinstance(obj, type)
        and obj is not channels.FieldChannelMixin
        and issubclass(obj, channels.FieldChannelMixin)
    )


original_chart_to_dict = TopLevelMixin.to_dict


def chart_to_dict(self, *args, **kwargs):
    """Ignore warnings about data not being a sublcass of altair.Data"""
    with warnings.catch_warnings():
        warnings.simplefilter("ignore", UserWarning)
        result = original_chart_to_dict(self, *args, **kwargs)
    return result


original_encode = _EncodingMixin.encode

channel_key_to_cls = {
    name.lower(): cls for name, cls in inspect.getmembers(channels, is_channel_cls)
}


@curry
def _wrap_in_channel_cls(name, obj):
    if not isinstance(obj, FieldChannelMixin) and isinstance(
        obj, AltairEncodingChannelHook
    ):
        return channel_key_to_cls[name](obj)
    return obj


def encode(self, *args, **kwargs):
    for name, obj in kwargs.items():
        if isinstance(obj, AltairEncodingChannelHook):
            # shorthands passed directly are wrapped with encoding channel constructor
            # channel's __init__ will handle resolution
            kwargs[name] = _wrap_in_channel_cls(name, obj)
        elif isinstance(obj, (list, tuple)):
            kwargs[name] = list(map(_wrap_in_channel_cls(name), obj))
    return original_encode(self, *args, **kwargs)


def _resolve_repeat_objs(objs: list) -> list:
    if objs is Undefined:
        return objs
    result = []
    for obj in objs:
        if isinstance(obj, AltairEncodingChannelHook):
            result.append(obj.encoding_fields()["field"])
        else:
            result.append(obj)
    return result


original_repeat = TopLevelMixin.repeat


def repeat(
    self,
    repeat=Undefined,
    row=Undefined,
    column=Undefined,
    layer=Undefined,
    columns=Undefined,
    **kwargs,
):
    repeat = _resolve_repeat_objs(repeat)
    row = _resolve_repeat_objs(row)
    column = _resolve_repeat_objs(column)
    layer = _resolve_repeat_objs(layer)
    return original_repeat(
        self,
        repeat=repeat,
        row=row,
        column=column,
        layer=layer,
        columns=columns,
        **kwargs,
    )


original_facet = _EncodingMixin.facet


def _resolve_facet_def(obj, cls):
    if isinstance(obj, AltairEncodingChannelHook):
        fields = obj.encoding_fields()
        return cls.from_dict(fields)
    return obj


def facet(
    self,
    facet=Undefined,
    row=Undefined,
    column=Undefined,
    data=Undefined,
    columns=Undefined,
    **kwargs,
):
    facet = _resolve_facet_def(facet, Facet)
    row = _resolve_facet_def(row, Row)
    column = _resolve_facet_def(column, Column)
    return original_facet(
        self, facet=facet, row=row, column=column, data=data, columns=columns, **kwargs
    )


def _update_nested(a, b):
    """Like Altair's update_nested, but convert everything to dicts beforehand."""
    for k, v in b.items():
        if isinstance(v, SchemaBase):
            b[k] = v.to_dict()
    return update_nested(a, b)


def build_channel_init(original_channel_init):
    def channel_init(self, *args, **kwargs):
        if args:
            shorthand, *args = args
            if isinstance(shorthand, AltairEncodingChannelHook):
                fields = shorthand.encoding_fields(self.__class__)
                kwargs = _update_nested(fields, kwargs)
            else:
                kwargs["shorthand"] = shorthand
        if "sort" in kwargs and isinstance(kwargs["sort"], AltairEncodingChannelHook):
            kwargs["sort"] = EncodingSortField(
                field=kwargs["sort"].encoding_fields()["field"]
            )
        return original_channel_init(self, *args, **kwargs)

    return channel_init


original_encoding_sort_field_init = EncodingSortField.__init__


def encoding_sort_field_init(
    self, field=Undefined, op=Undefined, order=Undefined, **kwds
):
    if isinstance(field, AltairEncodingChannelHook):
        field = field.encoding_fields()["field"]
    return original_encoding_sort_field_init(
        self, field=field, op=op, order=order, *kwds
    )


original_repr_mimebundle = TopLevelMixin._repr_mimebundle_


def _prep_data(data: List[dict]) -> List[dict]:
    df = pd.DataFrame(data)

    # dates are not auto-converted to Pandas datetime and so are not sanitized
    for col in df.columns:
        if df[col].apply(lambda x: isinstance(x, datetime.date)).any():
            df[col] = pd.to_datetime(df[col])

    return to_values(df)


type_to_encoding_type_metrics = {
    "bool": "ordinal",
    "date": "temporal",
    "datetime": "temporal",
    "int": "quantitative",
    "float": "quantitative",
    "str": "nominal",
}

type_to_encoding_type_dimensions = {
    "bool": "ordinal",
    "date": "temporal",
    "datetime": "temporal",
    "int": "ordinal",
    "float": "quantitative",
    "str": "nominal",
}


def get_default_encoding_type(info: DisplayInfo):
    """Get a default encoding type based on the column's DisplayInfo"""
    if info.kind == "metric":
        return type_to_encoding_type_metrics[info.type]
    return type_to_encoding_type_dimensions[info.type]


def build_default_tooltip(
    query: Query, info: Dict[str, DisplayInfo], locale: str
) -> List[dict]:
    result = []
    for request in chain(query.dimensions, query.metrics):
        field = request.name
        title = info[field].display_name
        format = format_config_to_d3_format(info[field].format, locale=locale)
        item = {
            "field": field,
            "title": title,
            "type": get_default_encoding_type(info[field]),
        }
        if format is not None:
            item["format"] = format
        result.append(item)
    return result


def render_self(self):
    """Extract and execute the query implied by the chart spec, embed the data, set
    formats, axis/legend titles etc.
    """
    self = self.copy(deep=True)  # don't mutate the original
    currencies = set()
    currency = None
    model = None

    for unit, data in self._iterunits():
        if isinstance(data, DictumData):
            query: Query = unit._query()
            result = data.execute(query)

            model = data.project.model
            currencies |= model.get_currencies_for_query(query)

            if "repeat" in unit._kwds:
                # repeat (unlike facet) expects the data to be in the unit
                unit.spec.data = _prep_data(result.data)
            else:
                unit.data = _prep_data(result.data)

            for channel in unit._iterchannels():
                if isinstance(channel, list) or isinstance(
                    channel.shorthand, RepeatRef
                ):
                    continue  # skip repeat channels

                if channel.field not in result.display_info:
                    continue  # skip literal channels

                info = result.display_info[channel.field]
                title = {"title": info.display_name}
                fmt = format_config_to_d3_format(info.format, model.locale)
                if fmt is not None:
                    title["format"] = fmt

                fields = {k: title for k in ["axis", "legend", "header"]}
                fields["type"] = get_default_encoding_type(info)
                fields = _update_nested(fields, channel.to_dict())
                fields = filter_fields(channel.__class__, fields)
                if (
                    fields["type"] == "temporal"
                    and "timeUnit" not in fields
                    and info.altair_time_unit is not None
                ):
                    fields["timeUnit"] = info.altair_time_unit
                channel._kwds.update(fields)
                if isinstance(channel, Text) and fmt is not None:
                    channel.format = fmt

        # add default tooltip if necessary
        tooltip_unit = unit
        if "spec" in unit._kwds:
            tooltip_unit = unit.spec
        if tooltip_unit.encoding.tooltip is Undefined:
            tooltip_unit.encoding.tooltip = build_default_tooltip(
                query, result.display_info, model.locale
            )

    if len(currencies) == 1:
        # if there's more than one currency per query, display no currency
        currency = currencies.pop()
    if model is not None:
        renderers.set_embed_options(
            formatLocale=cldr_locale_to_d3_number(model.locale, currency),
            timeFormatLocale=cldr_locale_to_d3_time(model.locale),
        )
    return self


def rendered_dict(self):
    return self._render_self().to_dict()


def repr_mimebundle(self, *args, **kwargs):
    self = self._render_self()
    return original_repr_mimebundle(self, *args, **kwargs)


def inject(method: callable, name: str, *classes):
    """Add a method into multiple classes"""
    for cls in classes:
        setattr(cls, name, method)


def iterattrs_with_names(self):
    yield from (
        (name, attr) for name, attr in self._kwds.items() if attr is not Undefined
    )


def iterattrs(self):
    """Iterate over non-undefined attributes"""
    yield from (attr for _, attr in iterattrs_with_names(self))


_items = ["layer", "hconcat", "vconcat", "concat"]


def iteritems(self):
    """Iterate over the immediately-embedded specs."""
    for key in _items:
        if hasattr(self, key):
            yield from getattr(self, key)


def iterunits(self, data=None):
    """
    Iterate over all the embedded lowest-level specs,
    returning a tuple of spec, data
    """
    if "mark" in self._kwds:
        yield self, self.data if data is None else data
        return
    if "facet" in self._kwds:
        yield self, self.data if data is None else data
        return
    if "repeat" in self._kwds:
        yield self, self.spec.data if data is None else data
    for item in self._iteritems():
        yield from item._iterunits(self.data if self.data is not Undefined else data)


def iterchannels(self):
    if hasattr(self, "spec"):
        yield from self.spec._iterchannels()
        if "facet" in self._kwds:
            if isinstance(self.facet, Facet):
                yield self.facet
            else:
                for attr in self.facet._iterattrs():
                    if isinstance(attr, (Row, Column)):
                        yield attr
        elif "repeat" in self._kwds:
            if isinstance(self._kwds["repeat"], list):
                yield self._kwds["repeat"]
            else:
                for attr in self._kwds["repeat"]._iterattrs():
                    yield attr
        return
    if self.encoding is not Undefined:
        yield from self.encoding._iterattrs()
    for item in self._iteritems():
        yield from item._iterchannels()


def is_dictum_definition(req):
    if isinstance(req, FieldName):
        req = req.to_dict()
    return isinstance(req, str) and (
        req.startswith("metric:") or req.startswith("dimension:")
    )


def request_from_definition(defn: str):
    if isinstance(defn, FieldName):
        defn = defn.to_dict()
    type_, field = defn.split(":", maxsplit=1)
    if type_ == "metric":
        return compile_metric_request(field)
    if type_ == "dimension":
        return compile_dimension_request(field)
    raise ValueError(f"{defn} is not a valid Dictum definition.")


def request_from_field(field):
    if is_dictum_definition(field):
        return request_from_definition(field)


def requests_from_channel(channel):
    result = []
    if request := request_from_field(channel.field):
        result = [request]
        channel.field = request.name
    if "sort" in channel._kwds:
        if isinstance(channel.sort, dict):
            try:
                channel.sort = EncodingSortField.from_dict(channel.sort)
            except ValidationError:
                pass
        if isinstance(channel.sort, EncodingSortField) and is_dictum_definition(
            channel.sort.field
        ):
            request = request_from_field(channel.sort.field)
            result.append(request)
            channel.sort.field = request.name
    return result


def requests_from_list(items: List[str]):
    result = []
    for i, item in enumerate(items):
        if (req := request_from_field(item)) is not None:
            result.append(req)
            items[i] = req.name
    return result


def query(self):
    """Extract the implied semantic Query from a chart spec"""
    metrics = []
    dimensions = []

    def _add_requests(*reqs):
        for req in reqs:
            if isinstance(req, QueryMetricRequest):
                if req not in metrics:
                    metrics.append(req)
            elif req not in dimensions:
                dimensions.append(req)

    for channel in self._iterchannels():
        if isinstance(channel, channels.FieldChannelMixin):
            reqs = requests_from_channel(channel)
            _add_requests(*reqs)
        elif isinstance(channel, list) and all(isinstance(i, str) for i in channel):
            # for repeats
            reqs = requests_from_list(channel)
            _add_requests(*reqs)
        elif isinstance(channel, list):
            # for other listed channels
            for item in channel:
                reqs = requests_from_channel(item)
                _add_requests(*reqs)

    return Query(metrics=metrics, dimensions=dimensions)


units = (
    UnitSpec,
    TopLevelUnitSpec,
    FacetedUnitSpec,
    NormalizedFacetSpec,
    RepeatSpec,
    FacetChart,
    TopLevelRepeatSpec,
    LayerChart,
)


def monkeypatch_altair():
    TopLevelMixin.to_dict = chart_to_dict
    TopLevelMixin._render_self = render_self
    TopLevelMixin._rendered_dict = rendered_dict
    TopLevelMixin._repr_mimebundle_ = repr_mimebundle
    TopLevelMixin.repeat = repeat
    _EncodingMixin.encode = encode
    _EncodingMixin.facet = facet
    EncodingSortField.__init__ = encoding_sort_field_init
    for cls in channel_key_to_cls.values():
        cls.__init__ = build_channel_init(cls.__init__)
    inject(iterattrs_with_names, "_iterattrs_with_names", SchemaBase)
    inject(iterattrs, "_iterattrs", SchemaBase)
    inject(iteritems, "_iteritems", TopLevelMixin, NormalizedSpec)
    inject(iterunits, "_iterunits", TopLevelMixin, NormalizedSpec)
    inject(iterchannels, "_iterchannels", *units)
    inject(query, "_query", *units)



"""Same queries as with select, but in ql. We mainly care about everything working, the
query parsing tests are tested where ql itself is tested.
"""

from dictum_core.project import Project


def test_ql_query(project: Project):
    query = """
    select revenue.percent() as "Percent of Revenue"
    by invoice_date.year as Year, artist
    limit revenue.top(1) of (artist) within (invoice_date.year)
    """
    ql = project.ql(query)
    result = ql.df()
    assert result.shape[0] == 5

    fmt = ql.df(format=True)
    assert tuple(fmt.columns) == ("Year", "Artist", "Percent of Revenue")



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['dictum_core',
 'dictum_core.backends',
 'dictum_core.backends.mixins',
 'dictum_core.cli',
 'dictum_core.engine',
 'dictum_core.examples',
 'dictum_core.examples.chinook',
 'dictum_core.model',
 'dictum_core.model.expr',
 'dictum_core.project',
 'dictum_core.project.altair',
 'dictum_core.project.templates',
 'dictum_core.ql',
 'dictum_core.schema',
 'dictum_core.schema.analyses',
 'dictum_core.schema.model',
 'dictum_core.tests',
 'dictum_core.tests.test_end_to_end',
 'dictum_core.tests.test_engine',
 'dictum_core.tests.test_model',
 'dictum_core.tests.test_model.configs',
 'dictum_core.tests.test_project',
 'dictum_core.tests.test_project.test_altair',
 'dictum_core.tests.test_schema',
 'dictum_core.utils']

package_data = \
{'': ['*'],
 'dictum_core.cli': ['project_template/*',
                     'project_template/metrics/*',
                     'project_template/tables/*'],
 'dictum_core.examples.chinook': ['metrics/media/*',
                                  'metrics/orders/*',
                                  'metrics/people/*',
                                  'tables/*'],
 'dictum_core.tests.test_end_to_end': ['altair_output/*']}

install_requires = \
['Babel>=2.9.1,<3.0.0',
 'Jinja2>=3.0.1,<4.0.0',
 'PyYAML>=5.4.1,<6.0.0',
 'SQLAlchemy>=1.4.25,<2.0.0',
 'altair==4.2.0rc1',
 'jupyter>=1.0.0,<2.0.0',
 'lark>=0.11.3,<0.12.0',
 'pandas>=1.3.3,<2.0.0',
 'pydantic>=1.8.2,<2.0.0',
 'rich>=12.5.1,<13.0.0',
 'sqlparse>=0.4.2,<0.5.0',
 'typer>=0.6.1,<0.7.0']

entry_points = \
{'console_scripts': ['dictum = dictum_core.cli:app']}

setup_kwargs = {
    'name': 'dictum-core',
    'version': '0.1.7',
    'description': 'Core library for Dictum',
    'long_description': None,
    'author': 'Mikhail Akimov',
    'author_email': None,
    'maintainer': None,
    'maintainer_email': None,
    'url': None,
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'entry_points': entry_points,
    'python_requires': '>=3.8,<4',
}


setup(**setup_kwargs)



from unittest import mock

import pytest

from dictum_core.backends.base import ExpressionTransformer
from dictum_core.backends.sql_alchemy import SQLAlchemyBackend


@pytest.fixture(scope="module", autouse=True)
def patch_sqlalchemy_url():
    with mock.patch(
        "dictum_core.backends.sql_alchemy.SQLAlchemyBackend.url", "sqlite://"
    ):
        yield


def test_compiler_case():
    compiler = mock.MagicMock()
    transformer = ExpressionTransformer(compiler)
    transformer.case([1, 2, 3, 4])
    compiler.case.assert_called_once_with([[1, 2], [3, 4]], else_=None)

    transformer.case([1, 2, 3, 4, 5])
    compiler.case.assert_called_with([[1, 2], [3, 4]], else_=5)


def test_sqlalchemy_no_default_schema():
    backend = SQLAlchemyBackend()
    with mock.patch("dictum_core.backends.sql_alchemy.Table") as Table:
        backend.table("table")
        assert Table.call_args[1]["schema"] is None


def test_sqlalchemy_default_schema():
    backend = SQLAlchemyBackend(default_schema="test")
    with mock.patch("dictum_core.backends.sql_alchemy.Table") as Table:
        backend.table("table")
        assert Table.call_args[1]["schema"] == "test"


def test_sqlalchemy_override_default_schema():
    backend = SQLAlchemyBackend(default_schema="test")
    with mock.patch("dictum_core.backends.sql_alchemy.Table") as Table:
        backend.table("table", schema="override")
        assert Table.call_args[1]["schema"] == "override"



from babel import dates, numbers

import dictum_core.project.altair.locale


def get_default_format_for_kind(kind: str, locale: str) -> str:
    data = dictum_core.project.altair.locale.load_locale(locale)
    if kind == "date":
        return ldml_date_to_d3_time_format(data.date_formats["short"].pattern)
    if kind == "datetime":
        short_date = data.date_formats["short"].pattern
        short_time = data.time_formats["short"].pattern
        datetime = data.datetime_formats["short"].format(short_time, short_date)
        return ldml_date_to_d3_time_format(datetime)
    if kind == "currency":
        return ldml_number_to_d3_format(data.currency_formats["standard"].pattern)
    if kind == "percent":
        return ldml_number_to_d3_format(data.percent_formats[None].pattern)
    if kind in {"number", "decimal"}:
        return ldml_number_to_d3_format(data.decimal_formats[None].pattern)


def ldml_number_to_d3_format(pattern: str) -> str:
    """Convert CLDR number format pattern to d3-format specifier.

    Since CLDR only supports zero padding, fill and align aren't used.
    See tests for comments on what can't be supported.

    [[fill]align][sign][symbol][0][width][,][.precision][~][type]

    e - exponent notation.
    f - fixed point notation.
    g - either decimal or exponent notation, rounded to significant digits.
    r - decimal notation, rounded to significant digits.
    s - decimal notation with an SI prefix, rounded to significant digits.
    % - multiply by 100, and then decimal notation with a percent sign.
    p - multiply by 100, round to significant digits, and then decimal notation with a
        percent sign.
    b - binary notation, rounded to integer.
    o - octal notation, rounded to integer.
    d - decimal notation, rounded to integer.
    x - hexadecimal notation, using lower-case letters, rounded to integer.
    X - hexadecimal notation, using upper-case letters, rounded to integer.
    c - character data, for a string of text.
    """
    pat: numbers.NumberPattern = numbers.parse_pattern(pattern)

    sign = ""
    if pat.prefix[0] == "+":
        sign = "+"
    elif pat.prefix[0] == " ":
        sign = " "
    elif pat.prefix[1] == "(":
        sign = "("

    symbol = ""
    if "¤" in "".join(pat.prefix + pat.suffix):
        symbol = "$"

    zero = ""
    if pat.int_prec[0] > 0:
        zero = "0"

    width = ""
    if pat.int_prec[0] > 0:
        width = f"{pat.int_prec[0]}"

    comma = ""
    if pat.grouping != (1000, 1000):  # that's how no grouping is reprd in babel
        comma = ","

    tilde = ""
    if pat.frac_prec[0] != pat.frac_prec[1]:
        tilde = "~"

    precision = max(pat.frac_prec)

    type_ = ""
    if "%" in "".join(pat.suffix + pat.prefix) and pat.scale == 2:
        type_ = "%"
    elif pat.exp_prec is not None:
        type_ = "e"
    elif "@" in pattern:
        # precision doesn't mean the same thing for r type
        if pat.int_prec[0] != pat.int_prec[1]:
            tilde = "~"
        precision = pat.int_prec[1]
        zero, width = "", ""
        type_ = "r"
    elif precision == 0:
        type_ = "d"
    else:
        type_ = "f"

    precision = f".{precision}"

    if type_ == "d":
        precision = ""

    return f"{sign}{symbol}{zero}{width}{comma}{precision}{tilde}{type_}"


class DateFormatMapper:
    mapping = {
        # Year
        "Y": "%-y",
        "YY": "%y",
        "YYYY": "%Y",
        "y": "%-Y",
        "yyyy": "%G",
        "yy": "%g",
        # Quarter
        "Q": "%-q",
        "QQ": "%q",
        "QQQ": "%q",
        "q": "%-q",
        "qq": "%q",
        "qqq": "%q",
        # Month
        "M": "%-m",
        "MM": "%m",
        "MMM": "%b",
        "MMMM": "%B",
        "MMMMM": "%b",
        "L": "%-m",
        "LL": "%m",
        "LLL": "%b",
        "LLLL": "%B",
        "LLLLL": "%b",
        # Week
        "w": "%-V",
        "ww": "%V",
        "W": "%-W",
        "WW": "%W",
        # Day
        "d": "%-d",
        "dd": "%d",
        "D": "%-j",
        "DD": "%j",
        "DDD": "%j",
        # lala
        "E": "%a",  # abbreviated
        "EE": "%a",
        "EEE": "%a",
        "EEEE": "%A",  # full
        "EEEEE": "%a",  # narrow
        # month name
        "MMM": "%b",
        "MMMM": "%B",
        "MMMMM": "%b",
        "LLL": "%b",
        "LLLL": "%B",
        "LLLLL": "%b",
        # day of month
        "d": "%-d",
        "dd": "%d",
        "D": "%-j",
        "DD": "%j",
        "DDD": "%j",
        # day of week
        "E": "%a",
        "EE": "%a",
        "EEE": "%a",
        "EEEE": "%A",
        "e": "%a",
        "ee": "%a",
        "eee": "%a",
        "eeee": "%A",
        # Time period (AM/PM)
        "a": "%p",
        # Hour
        "h": "%-I",
        "hh": "%I",
        "H": "%-H",
        "HH": "%H",
        "K": "%-I",
        "KK": "%I",
        "k": "%-H",
        "kk": "%H",
        # Minute
        "m": "%-m",
        "mm": "%m",
        # Second
        "s": "%-S",
        "ss": "%S",
        "S": "%-S",
        "SS": "%S",
        # Timezone
        "z": "%Z",
        "zz": "%Z",
        "zzz": "%Z",
        "zzzz": "%Z",
        "Z": "%Z",
        "ZZ": "%Z",
        "ZZZ": "%Z",
        "ZZZZ": "%Z",
    }

    def __getitem__(self, key: str):
        return self.mapping.get(key, "")


mapper = DateFormatMapper()


def ldml_date_to_d3_time_format(pattern: str) -> str:
    pat = dates.parse_pattern(pattern)
    return pat.format % mapper


def format_config_to_d3_format(config, locale):
    if config is None:
        return None

    if config.kind == "string":
        return None

    convert = ldml_number_to_d3_format
    if config.kind in {"date", "datetime"}:
        convert = ldml_date_to_d3_time_format

    pattern = config.pattern
    if config.skeleton is not None:
        _locale = dictum_core.project.altair.locale.load_locale(locale)
        format = _locale.datetime_skeletons.get(config.skeleton)
        if format is None:
            skel_key = dates.match_skeleton(config.skeleton, _locale.datetime_skeletons)
            if skel_key is not None:
                format = _locale.datetime_skeletons[skel_key]
        if format is not None:
            pattern = format.pattern

    if pattern is None:
        return get_default_format_for_kind(config.kind, locale)

    return convert(pattern)



import datetime

import pytest
from pandas import DataFrame
from pandas.api.types import is_datetime64_any_dtype

from dictum_core.backends.base import Backend
from dictum_core.engine import Column, Engine, RelationalQuery
from dictum_core.model import Model
from dictum_core.model.expr.parser import parse_expr
from dictum_core.schema import Query


@pytest.fixture(scope="module")
def compute_df(engine: Engine, backend: Backend):
    def compute(query: Query):
        computation = engine.get_computation(query)
        return DataFrame(computation.execute(backend).data)

    return compute


@pytest.fixture(scope="module")
def compute_results(chinook: Model, engine: Engine, backend: Backend):
    def compute(query: Query):
        computation = engine.get_computation(query)
        return computation.execute(backend).data

    return compute


def test_groupby(compute_df: callable):
    query = Query.parse_obj(
        {
            "metrics": [{"metric": {"id": "track_count"}}],
            "dimensions": [{"dimension": {"id": "genre"}}],
        }
    )
    df = compute_df(query)
    assert df[df["genre"] == "Rock"].iloc[0]["track_count"] == 1297


def test_filter(compute_df: callable):
    query = Query.parse_obj(
        {
            "metrics": [{"metric": {"id": "items_sold"}}],
            "filters": [
                {"id": "genre", "transforms": [{"id": "isin", "args": ["Rock"]}]},
                {
                    "id": "customer_country",
                    "transforms": [{"id": "isin", "args": ["USA"]}],
                },
            ],
        }
    )
    df = compute_df(query)
    assert df.iloc[0][0] == 157


@pytest.mark.skip
def test_convert_datetime(chinook: Model, connection):
    """Temporarily off, because not sure if this is needed. SQLite's datetime()
    returns a string, not a datetime (unlike a raw column), but everything is sent
    as a string to the frontend anyway. Only problem is the Python API, which we don't
    have yet.
    """
    q = Query(
        metrics=["items_sold"],
        dimensions=["invoice_date"],
    )
    comp = chinook.get_computation(q)
    df = connection.compute(comp)
    assert is_datetime64_any_dtype(df["invoice_date"].dtype)


def test_metric_not_measure(compute_df: callable):
    query = Query.parse_obj(
        {
            "metrics": [
                {"metric": {"id": "revenue"}},
                {"metric": {"id": "track_count"}},
                {"metric": {"id": "revenue_per_track"}},
            ]
        }
    )
    df = compute_df(query)
    assert next(df.round(2).itertuples()) == (0, 2328.6, 3503, 0.66)


def test_metric_with_groupby(compute_df: callable):
    query = Query.parse_obj(
        {
            "metrics": [{"metric": {"id": "arppu"}}, {"metric": {"id": "track_count"}}],
            "dimensions": [{"dimension": {"id": "genre"}}],
        }
    )
    df = compute_df(query)
    assert df.shape == (25, 3)


def test_multiple_facts(compute_df: callable):
    query = Query.parse_obj(
        {
            "metrics": [
                {"metric": {"id": "items_sold"}},
                {"metric": {"id": "track_count"}},
            ]
        }
    )
    df = compute_df(query)
    assert tuple(df.iloc[0]) == (2240, 3503)


def test_multiple_facts_dimensions(compute_df: callable):
    query = Query.parse_obj(
        {
            "metrics": [
                {"metric": {"id": "items_sold"}},
                {"metric": {"id": "track_count"}},
            ],
            "dimensions": [{"dimension": {"id": "genre"}}],
        }
    )
    df = compute_df(query)
    assert tuple(df[df["genre"] == "Rock"].iloc[0][["items_sold", "track_count"]]) == (
        835,
        1297,
    )


def test_if(compute_df: callable):
    """Test if() function and case when ... then ... else ... end constructs"""
    query = Query.parse_obj(
        {
            "metrics": [{"metric": {"id": "items_sold"}}],
            "dimensions": [
                {"dimension": {"id": "invoice_year"}},
                {"dimension": {"id": "leap_year"}},
            ],
        }
    )
    df = compute_df(query)
    assert df[df["leap_year"] == "Yes"].iloc[0]["invoice_year"] == 2012


def test_subquery_join(compute_df: callable):
    query = Query.parse_obj(
        {
            "metrics": [{"metric": {"id": "items_sold"}}],
            "dimensions": [{"dimension": {"id": "customer_orders_amount_10_bins"}}],
        }
    )
    df = compute_df(query)
    assert df.shape == (2, 2)
    assert df[df["customer_orders_amount_10_bins"] == 30].iloc[0].items_sold == 1708


@pytest.fixture(scope="module")
def compute(chinook: Model, backend: Backend):
    def computer(expr: str, type="datetime"):
        expr = parse_expr(expr)
        columns = [Column(name="value", expr=expr, type=type)]
        query = RelationalQuery(
            source=chinook.tables.get("media_types"),
            _groupby=columns,
            join_tree=[],
        )
        compiled = backend.compile_query(query)
        return str(backend.execute(compiled).iloc[0, 0])

    return computer


def test_datetrunc(compute):
    dt = "2021-12-19 14:05:38"

    def datetrunc(part: str):
        return compute(f"datetrunc('{part}', toDatetime('{dt}'))")

    assert datetrunc("year") == "2021-01-01 00:00:00"
    assert datetrunc("quarter") == "2021-10-01 00:00:00"
    assert datetrunc("month") == "2021-12-01 00:00:00"
    assert datetrunc("week") == "2021-12-13 00:00:00"
    assert (
        compute("datetrunc('week', toDatetime('2022-01-03'))") == "2022-01-03 00:00:00"
    )
    assert datetrunc("day") == "2021-12-19 00:00:00"
    assert datetrunc("hour") == "2021-12-19 14:00:00"
    assert datetrunc("minute") == "2021-12-19 14:05:00"
    assert datetrunc("second") == "2021-12-19 14:05:38"


def test_datepart(compute):
    """ISO states that the first week of the year is the week that contains the first
    Thursday of the year, or the week that contains January 4th. Not all database
    engines conform to this and it would be very tedious to implement that on the
    backend level, so we don't test for the edge cases (start and end of year) here.
    Just use whatever logic the RDBMS provides.
    """
    dt = "2021-12-19 14:05:38"

    def datepart(part: str):
        return int(compute(f"datepart('{part}', toDatetime('{dt}'))", type="int"))

    assert datepart("year") == 2021
    assert datepart("quarter") == 4
    assert compute("datepart('quarter', toDatetime('2021-12-31'))", "int") == "4"
    assert compute("datepart('quarter', toDatetime('2022-01-01'))", "int") == "1"
    assert datepart("month") == 12
    assert datepart("week") == 50
    assert datepart("day") == 19
    assert datepart("hour") == 14
    assert datepart("minute") == 5
    assert datepart("second") == 38

    # Monday is 1, Sunday is 7
    assert datepart("dow") == 7
    assert datepart("dayofweek") == 7
    assert compute("datepart('dow', toDatetime('2022-08-08'))", "int") == "1"
    assert compute("datepart('dayofweek', toDatetime('2022-08-08'))", "int") == "1"


def test_datediff(compute):
    def datediff(part, s, e):
        return int(
            compute(f"datediff('{part}', toDatetime('{s}'), toDatetime('{e}'))", "int")
        )

    assert datediff("year", "2021-12-31 23:59:59", "2022-01-01 00:00:00") == 1
    assert datediff("quarter", "2021-12-31 23:59:59", "2022-01-01 00:00:00") == 1
    assert datediff("month", "2021-12-31 23:59:59", "2022-01-01 00:00:00") == 1
    assert datediff("week", "2022-01-02 23:59:59", "2022-01-03 00:00:00") == 1
    assert datediff("day", "2021-12-31 23:59:59", "2022-01-01 00:00:00") == 1
    assert datediff("hour", "2021-12-31 23:59:59", "2022-01-01 00:00:00") == 1
    assert datediff("minute", "2021-12-31 23:59:59", "2022-01-01 00:00:00") == 1
    assert datediff("second", "2021-12-31 23:59:59", "2022-01-01 00:00:00") == 1

    assert datediff("year", "2021-01-01", "2021-12-31") == 0
    assert datediff("quarter", "2021-01-01", "2021-03-31") == 0
    assert datediff("month", "2021-01-01", "2021-01-31") == 0
    assert datediff("week", "2022-01-03", "2022-01-09") == 0
    assert datediff("day", "2021-12-31 00:00:00", "2021-12-31 23:59:59") == 0
    assert datediff("hour", "2021-12-31 04:00:00", "2021-12-31 04:59:59") == 0
    assert datediff("minute", "2021-12-31 04:59:00", "2021-12-31 04:59:59") == 0
    assert datediff("second", "2021-12-31 04:59:59", "2021-12-31 04:59:59") == 0


def test_date(compute_results: callable):
    query = Query.parse_obj(
        {
            "metrics": [{"metric": {"id": "revenue"}}],
            "dimensions": [{"dimension": {"id": "invoice_date"}}],
        }
    )
    results = compute_results(query)
    assert isinstance(results, list)
    assert isinstance(results[0], dict)
    assert isinstance(results[0]["invoice_date"], datetime.date)


def test_datetime(compute_results: callable):
    query = Query.parse_obj(
        {
            "metrics": [{"metric": {"id": "revenue"}}],
            "dimensions": [{"dimension": {"id": "invoice_datetime"}}],
        }
    )
    results = compute_results(query)
    assert isinstance(results, list)
    assert isinstance(results[0], dict)
    assert isinstance(results[0]["invoice_datetime"], datetime.datetime)


def test_division(compute_results: callable):
    """Test division semantics with an integer."""
    query = Query.parse_obj({"metrics": [{"metric": {"id": "pct_music_orders"}}]})
    results = compute_results(query)
    assert round(results[0]["pct_music_orders"], 2) == 0.95



from pathlib import Path

from dictum_core import Project
from dictum_core.schema.project import env


def generate():
    path = Path(__file__).parent
    env.data["CHINOOK_DATABASE"] = ""
    project = Project(path)
    with project.backend.engine.connect() as conn:
        conn.connection.executescript((path / "chinook.sqlite.sql").read_text())
    return project



from typing import Literal, Optional, Union

from pydantic import BaseModel, Field, root_validator

from dictum_core.schema import utils
from dictum_core.schema.model.type import Type

FormatKind = Literal[
    "number", "decimal", "percent", "currency", "date", "datetime", "string"
]


class FormatConfig(BaseModel):
    kind: FormatKind
    pattern: Optional[str]
    skeleton: Optional[str]
    currency: Optional[str]

    @root_validator(skip_on_failure=True)
    def validate_currency(cls, values):
        if values["kind"] != "currency":
            if values.get("currency") is not None:
                raise ValueError(
                    "'currency' format option is only valid with 'currency' format kind"
                )
            return values
        if values.get("currency") is None:
            raise ValueError(
                "'currency' formatting option is required if format kind is 'currency'"
            )
        if values["currency"] not in utils.currencies:
            raise ValueError(f"{values['currency']} is not a supported currency")
        return values

    @root_validator(skip_on_failure=True)
    def validate_pattern_skeleton(cls, values):
        pat = values.get("pattern")
        skel = values.get("skeleton")
        if pat is not None and skel is not None:
            raise ValueError("pattern and skeleton options are mutually exclusive")
        if skel is not None and values["kind"] not in {"date", "datetime"}:
            raise ValueError(
                "skeletons can only be used with date and datetime formats"
            )
        return values


Format = Union[FormatKind, FormatConfig]


class Formatted(BaseModel):

    format: Optional[Format] = Field()

    @staticmethod
    def get_default_format(type: Type) -> FormatConfig:
        if type in {"int", "float"}:
            return FormatConfig(kind="number")
        if type in {"date", "datetime"}:
            return FormatConfig(kind=type, pattern="short")
        if type in {"bool", "str"}:
            return FormatConfig(kind="string")
        raise ValueError

    @root_validator
    def set_default_format(cls, values):
        fmt = values.get("format")
        if fmt is None:
            values["format"] = cls.get_default_format(values.get("type", "float"))
        if isinstance(fmt, str):
            values["format"] = FormatConfig(kind=fmt)
        return values



from copy import deepcopy
from itertools import chain

from lark import Tree
from toolz import compose_left

from dictum_core import model, schema
from dictum_core.engine.aggregate_query_builder import AggregateQueryBuilder
from dictum_core.engine.computation import LiteralOrderItem, RelationalQuery
from dictum_core.engine.metrics import AddMetric, limit_transforms, transforms
from dictum_core.engine.operators import (
    FinalizeOperator,
    MaterializeOperator,
    MergeOperator,
)


def metric_expr(expr: Tree):
    expr = deepcopy(expr)
    for ref in expr.find_data("measure"):
        ref.data = "column"
        ref.children = [None, *ref.children]
    return expr


class Engine:
    def __init__(self, model: "model.Model"):
        self.model = model

    def suggest_dimensions(self, query: schema.Query):
        ...

    def suggest_measures(self, query: schema.Query):
        ...

    def get_range_computation(self, dimension_id: str) -> RelationalQuery:
        ...

    def get_values_computation(self, dimension_id: str) -> RelationalQuery:
        ...

    @staticmethod
    def validate_query(query: "schema.Query"):
        """Basic query validity checks"""
        # at least one metric is present
        if len(query.metrics) == 0:
            raise ValueError("You must request at least one metric")

        # no duplicate column names
        names = set()
        for request in chain(query.metrics, query.dimensions):
            name = request.name
            if name in names:
                raise ValueError(f"Duplicate column name in query: {name}")
            names.add(name)

        # all OF/WITHIN dimensions are also present in query.dimensions
        dimensions_digests = set(r.digest for r in query.dimensions)
        for request in query.metrics:
            for transform in request.metric.transforms:
                for item in chain(transform.of, transform.within):
                    exc = ValueError(
                        "All dimensions used in OF/WITHIN must also be present in the "
                        "query's dimension list.\n"
                        f"Metric: {request.render()}\nTransform: {transform.id}\n"
                        f"Dimension expression: {item.render()}"
                    )
                    if item.digest not in dimensions_digests:
                        raise exc

    def get_terminal(self, query: "schema.Query") -> MergeOperator:
        self.validate_query(query)

        builder = AggregateQueryBuilder(
            model=self.model, dimensions=query.dimensions, filters=query.filters
        )

        merge = MergeOperator(query=query)

        # add metrics
        adders = []
        for request in query.metrics:
            if len(request.metric.transforms) == 0:
                adders.append(AddMetric(request=request, builder=builder))
            elif len(request.metric.transforms) == 1:
                transform_id = request.metric.transforms[0].id
                adder = transforms.get(transform_id)
                if adder is None:
                    raise KeyError(f"Transform {transform_id} does not exist")
                adders.append(adder(request=request, builder=builder))

        for metric in query.limit:
            transform_id = metric.transforms[0].id
            adder = limit_transforms.get(transform_id)
            if adder is None:
                raise KeyError(f"Transform {transform_id} does not exist")
            adders.append(adder(metric=metric, builder=builder))

        return compose_left(*adders)(merge)

    def get_computation(self, query: schema.Query) -> MergeOperator:
        terminal = self.get_terminal(query)
        terminal.order = [LiteralOrderItem(r.digest, True) for r in query.dimensions]
        return FinalizeOperator(
            input=MaterializeOperator([terminal]),
            aliases={r.digest: r.name for r in query.metrics + query.dimensions},
        )



class ProjectMetricRequest:
    def total(self): ...
    def percent(self): ...
    def top(self, n: int): ...

class ProjectDimensionRequest:
    def year(self) -> "ProjectDimensionRequest": ...
    def quarter(self) -> "ProjectDimensionRequest": ...
    def month(self) -> "ProjectDimensionRequest": ...
    def week(self) -> "ProjectDimensionRequest": ...
    def day(self) -> "ProjectDimensionRequest": ...
    def date(self) -> "ProjectDimensionRequest": ...
    def hour(self) -> "ProjectDimensionRequest": ...
    def minute(self) -> "ProjectDimensionRequest": ...
    def second(self) -> "ProjectDimensionRequest": ...
    def datetrunc(self, part: str) -> "ProjectDimensionRequest": ...
    def datepart(self, part: str) -> "ProjectDimensionRequest": ...
    def last(self, n: int, part: str) -> "ProjectDimensionRequest": ...
    def step(self, size: int) -> "ProjectDimensionRequest": ...



from typing import Any, Optional

from pydantic import Field

from dictum_core.schema.id import ID
from dictum_core.schema.model.format import Formatted
from dictum_core.schema.model.type import Type


class Displayed(Formatted):
    id: ID
    name: str
    description: Optional[str]
    type: Type
    missing: Optional[Any]


class Calculation(Displayed):
    str_expr: str = Field(..., alias="expr")


class AggregateCalculation(Calculation):
    type: Type = "float"
    str_filter: Optional[str] = Field(alias="filter")
    str_time: Optional[str] = Field(alias="time")


class Measure(AggregateCalculation):
    metric: bool = False


class Metric(AggregateCalculation):
    table: Optional[str]  # this one is for metric-measures


class Dimension(Calculation):
    union: Optional[str]


class DetachedDimension(Dimension):
    """Just a dimension not defined on a table, the user has to explicitly
    specify which table it is.
    """

    table: str


class DimensionsUnion(Displayed):
    pass



from pathlib import Path

from lark import Lark, Token, Transformer, Tree

grammar = (Path(__file__).parent / "ql.lark").read_text()
ql = Lark(grammar, start="query")


def filter_tree(name: str, dimension: Tree, value):
    return Tree("filter", [dimension, Tree("call", [name, value])])


def call_scalar_children(name: str):
    def method(self, children: list):
        return Tree("scalar_transform", [name, *children])

    return method


class Preprocessor(Transformer):
    transform_aliases = {
        "not": "invert",
    }

    def identifier(self, children: list):
        return children[0]

    def IDENTIFIER(self, token: Token):
        return token.value

    def QUOTED_IDENTIFIER(self, token: Token):
        return token.value[1:-1]  # unquote

    def STRING(self, value: str):
        return value[1:-1]  # unquote

    def INTEGER(self, value: str):
        return int(value)

    def FLOAT(self, value: str):
        return float(value)

    def op(self, children: str):
        return children[0]

    def scalar_transform(self, children: list):
        id, *args = children
        id = self.transform_aliases.get(id, id)
        return Tree("scalar_transform", [id, *args])

    gt = call_scalar_children("gt")
    ge = call_scalar_children("ge")
    lt = call_scalar_children("lt")
    le = call_scalar_children("le")
    eq = call_scalar_children("eq")
    ne = call_scalar_children("ne")
    isnull = call_scalar_children("isnull")
    isnotnull = call_scalar_children("isnotnull")
    isin = call_scalar_children("isin")


pre = Preprocessor()


def parse_ql(query: str):
    return pre.transform(ql.parse(query))


dimension = Lark(grammar, start="dimension")


def parse_dimension(expr: str):
    """
    A separate function to parse string transform definitions during interactive use
    """
    return pre.transform(dimension.parse(expr))


dimension_request = Lark(grammar, start="dimension_request")


def parse_dimension_request(expr: str):
    return pre.transform(dimension_request.parse(expr))


metric = Lark(grammar, start="metric")


def parse_metric(expr: str):
    return pre.transform(metric.parse(expr))


metric_request = Lark(grammar, start="metric_request")


def parse_metric_request(expr: str):
    return pre.transform(metric_request.parse(expr))



from copy import deepcopy
from dataclasses import dataclass
from functools import cached_property
from typing import Any, Dict, List, Optional, Tuple

from lark import Transformer, Tree

import dictum_core.model
from dictum_core import schema
from dictum_core.model import utils
from dictum_core.model.expr import get_expr_kind, parse_expr
from dictum_core.utils import value_to_token


class ResolutionError(Exception):
    pass


@dataclass
class Displayed:
    id: str
    name: str
    description: str
    type: schema.Type
    format: Optional[schema.FormatConfig]
    missing: Optional[Any]

    def __post_init__(self):
        if isinstance(self.format, dict):
            self.format = schema.FormatConfig.parse_obj(self.format)


@dataclass
class Calculation(Displayed):
    """Parent class for measures and dimensions."""

    str_expr: str

    @cached_property
    def expr(self) -> Tree:
        raise NotImplementedError

    @cached_property
    def parsed_expr(self):
        return parse_expr(self.str_expr)

    @cached_property
    def expr_tree(self) -> str:
        return self.expr.pretty()

    @cached_property
    def kind(self) -> str:
        try:
            return get_expr_kind(self.parsed_expr)
        except ValueError as e:
            raise ValueError(
                f"Error in {self} expression {self.str_expr}: {e}"
            ) from None

    def check_references(self, path=tuple()):
        if self.id in path:
            raise RecursionError(f"Circular reference in {self}: {path}")
        self.check_measure_references(path + (self.id,))
        self.check_dimension_references(path + (self.id,))

    def check_measure_references(self, path):
        raise NotImplementedError

    def check_dimension_references(self, path):
        for ref in self.parsed_expr.find_data("dimension"):
            dimension = self.table.allowed_dimensions.get(ref.children[0])
            if dimension is None:
                raise KeyError(
                    f"{self} uses dimension {ref.children[0]}, but there's "
                    f"no unambiguous join path between {self.table} "
                    "and dimension's parent table"
                )
            dimension.check_references(path)

    def prefixed_expr(self, prefix: List[str]) -> Tree:
        return utils.prefixed_expr(self.expr, prefix)

    def prepare_expr(self, prefix: List[str]) -> Tree:
        return utils.prepare_expr(self.expr, prefix)

    def prepare_range_expr(self, base_path: List[str]) -> Tuple[Tree, Tree]:
        return (
            Tree("call", ["min", self.prepare_expr(base_path)]),
            Tree("call", ["max", self.prepare_expr(base_path)]),
        )

    def __str__(self):
        return f"{self.__class__.__name__}({self.id})"

    def __repr__(self):
        return str(self)

    def __hash__(self):
        return hash(self.id)

    def __eq__(self, other):
        return self.id == other.id


@dataclass(eq=False, repr=False)
class TableCalculation(Calculation):
    table: "dictum_core.model.Table"


class DimensionTransformer(Transformer):
    def __init__(
        self,
        table: "dictum_core.model.Table",
        measures: Dict[str, "Measure"],
        dimensions: Dict[str, "Dimension"],
        visit_tokens: bool = True,
    ) -> None:
        self._table = table
        self._measures = measures
        self._dimensions = dimensions
        super().__init__(visit_tokens=visit_tokens)

    def column(self, children: list):
        return Tree("column", [self._table.id, *children])

    def measure(self, children: list):
        measure_id = children[0]
        return Tree("column", [self._table.id, f"__subquery__{measure_id}", measure_id])

    def dimension(self, children: list):
        dimension_id = children[0]
        dimension = self._dimensions[dimension_id]
        path = self._table.dimension_join_paths.get(dimension_id, [])
        return dimension.prefixed_expr(path).children[0]


@dataclass(eq=False, repr=False)
class Dimension(TableCalculation):
    is_union: bool = False

    @cached_property
    def expr(self) -> Tree:
        self.check_references()
        transformer = DimensionTransformer(
            self.table, self.table.measure_backlinks, self.table.allowed_dimensions
        )
        expr = transformer.transform(self.parsed_expr)
        if self.missing is not None:
            expr = Tree(
                "expr",
                [
                    Tree(
                        "call",
                        [
                            "coalesce",
                            expr.children[0],
                            value_to_token(self.missing),
                        ],
                    )
                ],
            )
        return expr

    def check_measure_references(self, path=tuple()):
        for ref in self.parsed_expr.find_data("measure"):
            measure_id = ref.children[0]
            measure = self.table.measure_backlinks.get(measure_id).measures.get(
                measure_id
            )
            measure.check_references(path)


@dataclass
class TableFilter:
    table: "dictum_core.model.Table"
    str_expr: str

    @cached_property
    def expr(self) -> Tree:
        expr = parse_expr(self.str_expr)
        transformer = DimensionTransformer(
            table=self.table,
            measures=self.table.measure_backlinks,
            dimensions=self.table.allowed_dimensions,
        )
        return transformer.transform(expr)


@dataclass
class DimensionsUnion(Displayed):
    def __str__(self):
        return f"Union({self.id})"


class MeasureTransformer(Transformer):
    def __init__(
        self,
        table: "dictum_core.model.Table",
        measures: Dict[str, "Measure"],
        dimensions: Dict[str, "Dimension"],
        visit_tokens: bool = True,
    ) -> None:
        self._table = table
        self._measures = measures
        self._dimensions = dimensions
        super().__init__(visit_tokens=visit_tokens)

    def column(self, children: list):
        return Tree("column", [self._table.id, *children])

    def measure(self, children: list):
        ref_id = children[0]
        measure = self._measures[ref_id]
        expr = measure.expr.children[0]
        if measure.missing is not None:
            expr = Tree("call", ["coalesce", expr, value_to_token(measure.missing)])
        return expr

    def dimension(self, children: list):
        dimension_id = children[0]
        path = self._table.dimension_join_paths[dimension_id]
        return self._dimensions[dimension_id].prefixed_expr(path).children[0]


@dataclass(repr=False)
class Measure(TableCalculation):
    str_filter: Optional[str] = None
    str_time: Optional[str] = None

    def __post_init__(self):
        super().__post_init__()
        if self.kind != "aggregate":
            raise ValueError(
                f"Measures must be aggregate, {self} expression is not: {self.str_expr}"
            )

    @cached_property
    def expr(self) -> Tree:
        self.check_references()
        transformer = MeasureTransformer(
            self.table, self.table.measures, self.table.allowed_dimensions
        )
        return transformer.transform(self.parsed_expr)

    @cached_property
    def filter(self) -> Tree:
        if self.str_filter is None:
            return None
        self.check_references()
        transformer = DimensionTransformer(
            self.table, self.table.measure_backlinks, self.table.allowed_dimensions
        )
        return transformer.transform(parse_expr(self.str_filter))

    @property
    def time(self) -> Dimension:
        if self.str_time is None:
            raise ValueError(
                f"{self} doesn't have a time dimension specified so it "
                "can't be used with the built-in Time dimension"
            )

        return self.table.allowed_dimensions[self.str_time]

    @cached_property
    def dimensions(self):
        return self.table.allowed_dimensions.values()

    def check_measure_references(self, path=tuple()):
        for ref in self.parsed_expr.find_data("measure"):
            measure = self.table.measures.get(ref.children[0])
            measure.check_references(path)

    def __eq__(self, other: "Metric"):
        return self.id == other.id

    def __hash__(self):
        return hash(self.id)


class MetricTransformer(Transformer):
    def __init__(
        self,
        metrics: Dict[str, "Metric"],
        measures: Dict[str, "Measure"],
        visit_tokens: bool = True,
    ) -> None:
        self._metrics = metrics
        self._measures = measures
        super().__init__(visit_tokens=visit_tokens)

    def column(self, children: list):
        raise ValueError("Column references are not allowed in metrics")

    def dimension(self, children: list):
        raise ValueError("Dimension references are not allowed in metrics")

    def measure(self, children: list):
        ref_id = children[0]
        if ref_id in self._metrics:
            return self._metrics[ref_id].expr.children[0]
        if ref_id in self._measures:
            return Tree("measure", children)
        raise KeyError(f"reference {ref_id} not found")


@dataclass(repr=False)
class Metric(Calculation):
    model: "dictum_core.model.Model"
    is_measure: bool = False

    @classmethod
    def from_measure(
        cls, measure: Measure, model: "dictum_core.model.Model"
    ) -> "Metric":
        return cls(
            model=model,
            id=measure.id,
            name=measure.name,
            description=measure.description,
            str_expr=f"${measure.id}",
            type=measure.type,
            format=measure.format,
            missing=measure.missing,
            is_measure=True,
        )

    @cached_property
    def expr(self) -> Tree:
        metrics = self.model.metrics.copy()
        del metrics[self.id]
        transformer = MetricTransformer(metrics, self.model.measures)
        try:
            expr = transformer.transform(self.parsed_expr)
        except Exception as e:
            raise ResolutionError(f"Error resolving expression of {self}: {e}")
        if self.missing is not None:
            expr = Tree(
                "expr",
                [
                    Tree(
                        "call",
                        [
                            "coalesce",
                            expr.children[0],
                            value_to_token(self.missing),
                        ],
                    )
                ],
            )
        return expr

    @cached_property
    def merged_expr(self) -> Tree:
        """Same as expr, but measures are selected as column from the merged table"""
        expr = deepcopy(self.expr)
        for ref in expr.find_data("measure"):
            ref.data = "column"
            ref.children = [None, *ref.children]
        return expr

    @cached_property
    def measures(self) -> List[Measure]:
        result = []
        for ref in self.expr.find_data("measure"):
            result.append(self.model.measures.get(ref.children[0]))
        return result

    @cached_property
    def dimensions(self) -> Dict[str, Dimension]:
        return sorted(
            set.intersection(*(set(m.dimensions) for m in self.measures)),
            key=lambda x: x.name,
        )



reserved = {
    "Time",
    "Year",
    "Quarter",
    "Month",
    "Week",
    "Day",
    "Date",
    "Hour",
    "Minute",
    "Second",
}


class ID(str):
    @classmethod
    def __get_validators__(cls):
        yield cls.validate

    @classmethod
    def validate(cls, value: str):
        if not isinstance(value, str):
            raise TypeError(f"Invalid object ID: {value}. Object ID must be a string.")
        if not value.isidentifier():
            raise ValueError(
                f"Invalid object ID: {value}. "
                "Object ID must be a valid Python identifier."
            )
        if value.startswith("__"):
            raise ValueError(
                f"Invalid object ID: {value}. "
                "IDs starting with double underscore (__) "
                "are reserved for internal use."
            )
        if value in reserved:
            raise ValueError(
                f"Invalid object ID: {value}. IDs {reserved} are reserved."
            )
        return cls(value)



from pathlib import Path
import yaml
from dictum_core import Project

from typer.testing import CliRunner

from dictum_core.cli import app

runner = CliRunner()


def test_new(tmp_path: Path):
    test = tmp_path / "test"
    result = runner.invoke(app, args=["new"], input=f"{test}\nTest project\n1\n\n")
    assert result.exit_code == 0
    assert test.is_dir()
    assert (test / "project.yml").exists()

    assert yaml.safe_load((test / "profiles.yml").read_text()) == {
        "default_profile": "sqlite",
        "profiles": {"sqlite": {"parameters": {"database": None}, "type": "sqlite"}},
    }
    assert (test / "metrics").exists() and (test / "metrics").is_dir()
    assert (test / "tables").exists() and (test / "tables").is_dir()

    Project(test)  # generated project is at least valid
